/**
 * \file Basic_Consumer.hh
 * \author Scott F. H. Kaplan -- sfkaplan@cs.amherst.edu
 * \date November 2002
 * \brief Produces a single reference stream to canonical page IDs.
 *
 * A subclass of Consumer that emits a basic page-level reference
 * trace with no extra information about shared space or multiple
 * processes.  Canonical page numbers are emitted, however, so that
 * shared pages are not multiply counted.  The trace takes the
 * following form:
 *
 *   R 14fe3
 *
 * The first field is the ``type tag'':  A single character, either
 * 'R' or 'W', to indicate whether or not the reference is a read or a
 * write.  Note that this trace format does not distinguish between
 * kernel and user references.  Instruction references are considered
 * read operations.
 *
 * The second field is the page number: A hexidecimal value giving the
 * canonical page number being referenced.
 **/



// ===================================================================
// Avoid multiple inclusion.
#if !defined (_BASIC_CONSUMER_HH)
#define _BASIC_CONSUMER_HH
// ===================================================================



// ===================================================================
// INCLUDES

#include "Consumer.hh"
// ===================================================================



// ===================================================================
// CLASS DEFINITION

/**
 * \brief A subclass of Consumer that produces a simple reference
 *        trace.
 *
 * It relies on Consumer to reconcile virtual pages to canonical
 * pages, and then emits a string of references to those canonical
 * pages.
 **/
class Basic_Consumer : public Consumer {

public: // CONSTRUCTOR

  /**
   * \brief The constructor.
   * \param reference_reader The object that will read and parse the
   *        reference trace.
   * \param kernel_pathname The pathname of the kernel trace to be
   *        read.
   * \param output_file_pathname The pathname for the file that will
   *        contain the resultant reference stream.
   **/
  Basic_Consumer (Reference_Reader* const reference_reader,
		  const char* const kernel_pathname,
		  const char* const output_file_pathname);

  /**
   * \brief The deconstructor.
   *
   * Close the output file.
   **/
  virtual ~Basic_Consumer ();

protected: // VIRTUAL METHODS

  /** \brief Called before processing begins -- Does nothing. */
  virtual void initialize () {}

  /** \brief Called after processing ends -- Does nothing. */
  virtual void clean_up () {}

  /**
   * \brief Called after processing a reference record.
   *
   * Transform the reference type into one valid for the output
   * trace.  Lookup the canonical page underlying the referenced
   * virtual page.  Emit both to the output trace.
   **/
  virtual void act_on_reference_record ();

  /** \brief Called after processing a kernel record -- Does
   *         nothing. */
  virtual void act_on_kernel_record () {}

protected: // DATA MEMBERS

  /** \brief The output file stream. */
  FILE* output_stream;

}; // class Basic_Consumer
// ===================================================================



// ===================================================================
#endif // _BASIC_CONSUMER_HH
// ===================================================================
