/**
 * \file Binary_Raw_Reader.cc
 * \author Scott F. H. Kaplan <sfkaplan@cs.amherst.edu>
 * \date January 2003
 * \brief Code for a class that reads and parses binary raw reference
 * traces.
 *
 * A subclass of Reference_Reader that can read the binary raw
 * reference traces directly produced by Laplace.  The data from each
 * record is then transformed appropriately for use by Laplace-merge.
 */



// ===================================================================
// INCLUDES

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include "Binary_Raw_Reader.hh"
#include "types-and-constants.hh"
// ===================================================================



// ===================================================================
// TYPES
// Definitions of types for each field in the raw reference trace.
// This is adapted from tracer.cc in Laplace itself.

/**
 * \brief The binary format of a raw reference record.
 */
struct binary_record_s {

  tag_t tag;
  timestamp_t timestamp;
  unsigned char length;
  context_ID_t context_ID;
  virtual_address_t virtual_address;

};
// ===================================================================



// ===================================================================
Binary_Raw_Reader::Binary_Raw_Reader
(const char* const reference_trace_pathname) :
  Reference_Reader(reference_trace_pathname) {

  // Reset the clocks to zero.
  instruction_time = 0;
  reference_time = 0;

} // Binary_Raw_Reader
// ===================================================================



// ===================================================================
void
Binary_Raw_Reader::read (reference_record_s* const reference_record) {

  // Attempt to read the next record of the reference trace.
  binary_record_s binary_record;
  size_t read_result = fread(&binary_record,
			     sizeof(binary_record),
			     1,
			     reference_stream);

  // Did the read fail?
  if (read_result != 1) {

    // Distinguish between an EOF and other errors, but abort
    // processing in either case.
    if (feof(reference_stream)) {
      fprintf(stderr,
	      "Binary_Raw_Reader::read(): Premature EOF\n");
    } else {
      fprintf(stderr,
	      "Binary_Raw_Reader::read(): Read error\n");
    }
    exit(1);
  }

  // Based on the type of reference, update some data and determine
  // what result to return.
  timestamp_t current_instruction_time = instruction_time;
  timestamp_t current_reference_time = reference_time;
  switch (binary_record.tag) {

  case TAG_USER_INSTRUCTION_FETCH:
  case TAG_KERNEL_INSTRUCTION_FETCH:

    // Tick the instruction clock to reflect that _after_ this event,
    // instruction time will have advanced.
    instruction_time++;

    // Fall through to read and parse the rest of the record...

  case TAG_USER_READ:
  case TAG_KERNEL_READ:
  case TAG_USER_WRITE:
  case TAG_KERNEL_WRITE:

    // Tick the reference clock to reflect that _after_ this event,
    // reference time will have advanced.
    reference_time++;

    // Copy elements of the record into the given structure that
    // require no processing.
    reference_record->tag = binary_record.tag;
    reference_record->cycle_timestamp = binary_record.timestamp;
    reference_record->length = binary_record.length;
    reference_record->virtual_page.context_ID =
      binary_record.context_ID;
    reference_record->virtual_address = binary_record.virtual_address;

    // Place the current instruction and reference times into the
    // structure.
    reference_record->instruction_timestamp = current_instruction_time;
    reference_record->reference_timestamp = current_reference_time;

    // Comput the page ID by isolating the upper bits of the virtual
    // address and recording that value in the structure.
    reference_record->virtual_page.page_ID =
      binary_record.virtual_address >> PAGE_SIZE_ORDER;

    break;

  case TAG_KERNEL_EVENT:

    // Read the remainder of the record.  We care only about the
    // timestamp.
    reference_record->tag = binary_record.tag;
    reference_record->cycle_timestamp = binary_record.timestamp;
    reference_record->instruction_timestamp = current_instruction_time;
    reference_record->reference_timestamp = current_reference_time;
    reference_record->length = 0;
    reference_record->virtual_page.context_ID = 0;
    reference_record->virtual_page.page_ID = 0;
    reference_record->virtual_address = 0;
    break;

  case TAG_END_OF_TRACE:

    // Nothing to do.
    break;

  default:
    fprintf(stderr,
	    "Binary_Raw_Reader::read(): Invalid reference tag %c(%d)\n",
	    reference_record->tag,
	    (int)reference_record->tag);
    exit(1);

  }

} // read
// ===================================================================
