/**
 * \file Consumer.cc
 * \author Scott F. H. Kaplan <sfkaplan@cs.amherst.edu>
 * \date November 2002
 * \brief Code of base class for reading traces and rebuilding system
 *        state.
 *
 * A class used to read both a kernel and reference trace and, with
 * each record, update data structures that reflect the state of the
 * processes and the memory.
 *
 * Note that this is a PURE VIRTUAL CLASS: It does not define the
 * initialize(), act_on_kernel_record(), act_on_reference_record(),
 * and clean_up() methods.  By defining these methods, a subclass can
 * choose to maintain extra data structures or to produce output
 * before processing, at the processing of each record, and after
 * processing.
 **/



// ===================================================================
// INCLUDES

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "Consumer.hh"
// ===================================================================



// ===================================================================
Consumer::Consumer (Reference_Reader* const reference_reader,
		    const char* const kernel_pathname) {

  // Hold a pointer to the reference reader.
  this->reference_reader = reference_reader;

  // Attempt to open the kernel trace file.
  kernel_stream = fopen(kernel_pathname, "r");

  // Did the kernel trace open correctly?
  if (kernel_stream == NULL) {

    // No.  Emit the error and exit.
    perror("Failed opening kernel trace");
    exit(1);

  }

  // Initialize some data members.
  number_reference_records = 0;
  number_kernel_records = 0;
  next_canonical_page_ID = 0;
  canonical_zero_page = get_new_canonical_page(TYPE_ZERO_PAGE);
  next_process_ID = KERNEL_PROCESS_ID + 1;

  // Create space to hold one reference record and one kernel record.
  reference_record = new reference_record_s;
  kernel_record = new kernel_record_s;
  if ((reference_record == 0) || (kernel_record == 0)) {
    fprintf(stderr, "Consumer(): Unable to allocate record space\n");
    exit(1);
  }

  // Create a task object for the initial kernel task as well as a
  // process object for the ``kernel process''.
  kernel_task = allocate_new_task();
  Process* kernel_process = allocate_new_process();

  // Give each an ID of 0, initialize the other fields, and assign the
  // thread to the process.
  kernel_task->task_ID = 0;
  kernel_task->process = kernel_process;
  kernel_task->parent_task = NULL;
  kernel_process->process_ID = KERNEL_PROCESS_ID;
  kernel_process->task_list.push_front(kernel_task);
  kernel_process->image_file_ID.inode_ID = 0;
  kernel_process->image_file_ID.major = 0;
  kernel_process->image_file_ID.minor = 0;
  strcpy(kernel_process->image_pathname, "***KERNEL***");

  // Create the virtual->canonical map (the context itself) and
  // associate it with the kernel process.
  virtual_to_canonical_map_t* kernel_context_map =
    new virtual_to_canonical_map_t;
  if (kernel_context_map == 0) {
    fprintf(stderr, "Consumer(): Unable to allocate kernel context map\n");
    exit(1);
  }
  kernel_process->virtual_to_canonical_map = kernel_context_map;
  kernel_process->context_ID = KERNEL_CONTEXT;

  // Insert them into both the task_ID- and context_ID-keyed tables.
  task_ID_to_Task_map[kernel_task->task_ID] = kernel_task;
  context_ID_to_Process_map[kernel_process->context_ID] =
    kernel_process;

  // Track the currently active task, which is initially the kernel task.
  active_task = kernel_task;

  // Nullify the scheduled task, as there isn't one yet.
  scheduled_task = NULL;

  // Initialize the single canonical buffer cache page.  (Reminder:
  // This is a cheap trick until we really track buffer use.  See
  // Consumer.hh for further notes.)
  canonical_buffer_cache_page.page = 0;
  canonical_buffer_cache_page.reference_count = 0;
  canonical_buffer_cache_page.type = TYPE_BUFFER_CACHE;

} // Consumer::Consumer
// ===================================================================



// ===================================================================
Consumer::~Consumer () {

  // WARNING: We do little here to clean up this object.  We (lazily)
  // assume that when this object dies, the process is complete
  // anyway.

  // Attempt to close the kernel trace file.
  int kernel_result = fclose(kernel_stream);

  // Did the kernel trace close correctly?
  if (kernel_result == EOF) {

    // No.  Emit the error and exit.
    perror("Failed closing kernel trace");
    exit(1);

  }

} // Consumer::~Consumer
// ===================================================================



// ===================================================================
// Perform the merging.
void
Consumer::merge () {

  // Initialize any output.
  initialize();

  // Perform the initial synchronizing phase, getting the reference
  // and kernel traces in sync.
  synchronize();

  // Set the first reference record (marking the last kernel event) as
  // the start of scheduled time for the kernel task.
  kernel_task->
    mark_quanta_start(reference_record->cycle_timestamp,
		      reference_record->instruction_timestamp,
		      reference_record->reference_timestamp);

  // Read the first reference record to be processed.  Since this is
  // the beginning of virtual time passage, set this as the start of
  // the scheduling quanta (as far as the traces indicate) for the
  // active task.
  reference_reader->read(reference_record);

  // Merge the traces until the last reference has been processed.
  while (reference_record->tag != TAG_END_OF_TRACE) {

    // Update the time that has passed between the last reference
    // record and this one for the active task.
    active_task->mark_time(reference_record->cycle_timestamp,
			   reference_record->instruction_timestamp,
			   reference_record->reference_timestamp);

    // Does the reference record indicate that there's a kernel record
    // to read and process?
    if (reference_record->tag == TAG_KERNEL_EVENT) {

      // It does.  Read the kernel record, update the maps with this
      // kernel record, and act on the record (that is, output and
      // other bookkeeping).
      read_kernel_record();
      update_with_kernel_record();
      act_on_kernel_record();

    } else {

      // No, this is just a reference record.  Update the maps with
      // this reference record, and act on the record (that is, output
      // and other bookkeeping).
      update_with_reference_record();
      act_on_reference_record();

    }

    // Read the next reference record.
    reference_reader->read(reference_record);

  } // Reference record reading loop

  // Complete processing by finishing output and other bookkeeping.
  clean_up();

} // Consumer::merge
// ===================================================================



// ===================================================================
void
Consumer::synchronize () {

  // Read the first reference record which must be a kernel event
  // record that indicates the time of the _last_ kernel event
  // preceeding references.
  reference_reader->read(reference_record);
  if (reference_record->tag != TAG_KERNEL_EVENT) {
    fprintf(stderr, "Consumer::synchronize(): ");
    fprintf(stderr, "First reference record not a kernel event\n");
    exit(1);
  }
  timestamp_t last_kernel_event_time =
    reference_record->cycle_timestamp;

  // If the last kernel event time is 0, that means that reference
  // tracing began during system startup, _before_ the kernel was
  // loaded and any kernel events could be recorded.  Therefore, we
  // can skip synchronization.
  if (last_kernel_event_time == 0) {
    return;
  }

  // Start at the beginning of the kernel trace.  Read and process all
  // records concluding with the last one preceeding actual
  // references.  In this pass of the kernel trace, we seek only to
  // establish which tasks will be live at the time of the first
  // reference.
  do {

    // Read the next kernel record.
    read_kernel_record();

    // Sanity check: We shouldn't go too far in the kernel trace.
    if (kernel_record->cycle_timestamp > last_kernel_event_time) {
      fprintf(stderr, "Consumer::synchronize(): ");
      fprintf(stderr, "Went too far in kernel trace\n");
      exit(1);
    }

    // Determine what to do based on the record type.
    switch (kernel_record->tag) {

    case TAG_FORK:
      // Mark the new task as live.
      live_map[kernel_record->task_ID] = true;
      break;

    case TAG_EXIT:
      // Mark the terminated task as dead.  Check first, though,
      // to be sure that it was live and that the kernel trace is
      // sane.
      if (!live_map[kernel_record->task_ID]) {
	fprintf(stderr, "Consumer::synchronize(): ");
	fprintf(stderr,  "Exit record on dead task\n");
	fprintf(stderr,
		"cycle_timestamp = %qx\n",
		kernel_record->cycle_timestamp);
	exit(1);
      }
      live_map[kernel_record->task_ID] = false;
      break;

    } // switch-case on tag

  } while (kernel_record->cycle_timestamp != last_kernel_event_time);

  // Perform a second pass, where we will maintain, for each task, its
  // most recent executable image pathname -- something that will
  // continue throughout the merging.  When fork() records are
  // encountered, the task will adopt the most recent executable image
  // pathname of the parent, and that pathname will be emitted with
  // the new fork record.  Stop when the first reference time is
  // reached.

  // Process until the time of the first reference.
  rewind(kernel_stream);
  number_kernel_records = 0;
  do {

    // Read the next kernel record.
    read_kernel_record();

    // Sanity check: We shouldn't go too far in the kernel trace.
    if (kernel_record->cycle_timestamp > last_kernel_event_time) {
      fprintf(stderr, "Consumer::synchronize(): ");
      fprintf(stderr, "Went too far in kernel trace\n");
      exit(1);
    }

    // Update the tables based on this record and act on this record
    // (that is, output and bookkeeping).
    update_with_kernel_record();
    act_on_kernel_record();

  } while (kernel_record->cycle_timestamp != last_kernel_event_time);

} // synchronize
// ===================================================================



// ===================================================================
Task*
Consumer::safe_task_ID_to_task (const task_ID_t task_ID) {

  // Attempt lookup.
  task_ID_to_Task_map_t::iterator i = task_ID_to_Task_map.find(task_ID);

  // If the lookup failed then abort.
  if (i == task_ID_to_Task_map.end()) {
    fprintf(stderr, "Consumer::safe_task_ID_to_task(): ");
    fprintf(stderr, "Failed lookup\n");
    exit(1);
  }

  // Return the lookup result.
  return (*i).second;

}
// ===================================================================



// ===================================================================
Process*
Consumer::safe_context_ID_to_process (const context_ID_t context) {

  // Attempt lookup.
  context_ID_to_Process_map_t::iterator i =
    context_ID_to_Process_map.find(context);

  // If the lookup failed then abort.
  if (i == context_ID_to_Process_map.end()) {
    fprintf(stderr, "Consumer::safe_context_ID_to_process(): ");
    fprintf(stderr, "Failed lookup\n");
    exit(1);
  }

  // Return the lookup result.
  return (*i).second;

}
// ===================================================================



// ===================================================================
bool
Consumer::safe_live (const task_ID_t task_ID) {

  // Attempt lookup.
  live_map_t::iterator i = live_map.find(task_ID);

  // If the lookup failed then abort.
  if (i == live_map.end()) {
    fprintf(stderr, "safe_live(): Failed lookup\n");
    exit(1);
  }

  // Return the lookup result.
  return (*i).second;

}
// ===================================================================



// ===================================================================
virtual_to_canonical_map_t*
Consumer::safe_file_ID_to_V2C_map (const file_ID_s& file_ID) {

  virtual_to_canonical_map_t* map = 0;

  // Attempt lookup.
  file_ID_to_V2C_map_t::iterator i = file_ID_to_V2C_map.find(file_ID);

  // Did the lookup succeed?
  if (i == file_ID_to_V2C_map.end()) {

    // No.  Allocate a new map, and ensure that the allocation
    // succeeded.
    map = new virtual_to_canonical_map_t;
    if (map == 0) {
      fprintf(stderr, "safe_file_ID_to_V2C_map(): Failed allocation\n");
      exit(1);
    }

    // Associate the given file ID with this new map.
    file_ID_to_V2C_map[file_ID] = map;

  } else {

    // Yes, use the map found via lookup.
    map = (*i).second;

  }

  // Return the lookup result.
  return map;

}
// ===================================================================



// ===================================================================
virtual_to_canonical_map_t*
Consumer::safe_shm_ID_to_V2C_map (const shm_ID_t& shm_ID) {

  virtual_to_canonical_map_t* map = 0;

  // Attempt lookup.
  shm_ID_to_V2C_map_t::iterator i = shm_ID_to_V2C_map.find(shm_ID);

  // Did the lookup succeed?
  if (i == shm_ID_to_V2C_map.end()) {

    // No.  Allocate a new map, and ensure that the allocation
    // succeeded.
    map = new virtual_to_canonical_map_t;
    if (map == 0) {
      fprintf(stderr, "safe_shm_ID_to_V2C_map(): Failed allocation\n");
      exit(1);
    }

    // Associate the given shm ID with this new map.
    shm_ID_to_V2C_map[shm_ID] = map;

  } else {

    // Yes, use the map found via lookup.
    map = (*i).second;

  }

  // Return the lookup result.
  return map;

} // safe_shm_ID_to_V2C_map
// ===================================================================



// ===================================================================
void
Consumer::map_virtual_to_canonical
(virtual_to_canonical_map_t* const map,
 const virtual_page_ID_t virtual_page_ID,
 canonical_page_s* const canonical_page) {

  // The local page about to be assigned should exist.
  // created.
  if (canonical_page == NULL) {
    fprintf(stderr, "map_virtual_to_canonical(): Null canonical page\n");
    exit(1);
  }

  // Insert the mapping.
  (*map)[virtual_page_ID] = canonical_page;

} // map_virtual_to_canonical
// ===================================================================



// ===================================================================
void
Consumer::read_kernel_record () {

  // Attempt to read the next record of the kernel trace.
  char* read_result = fgets(buffer, BUFFER_SIZE, kernel_stream);

  // Did the read fail?
  if (read_result == NULL) {

    // Distinguish between an EOF and other errors, but abort
    // processing in either case.
    fprintf(stderr, "Consumer::read_kernel_record(): ");
    if (feof(kernel_stream)) {
      fprintf(stderr, "Premature EOF\n");
    } else {
      fprintf(stderr, "Read error\n");
    }
    exit(1);
  }

  // Parse the record type, which should be the first character of the
  // buffer.
  kernel_record->tag = buffer[0];

  // Attempt to parse the rest of the record based on its type.
  // Declare and assign some variables here to avoid compiler
  // complaints of skipping initialization in switch-case statement.
  unsigned int parse_result = 0;
  unsigned int items_to_parse = 0;
  virtual_address_t starting_virtual_address = 0;
  virtual_address_t ending_virtual_address = 0;
  offset_t length = 0;
  file_offset_t file_offset = 0;
  switch (kernel_record->tag) {

  case TAG_SCHEDULE:
  case TAG_ACCEPT:
  case TAG_EXIT:
    items_to_parse = 2;
    parse_result = sscanf(&buffer[2],
			  "%qx %hd",
			  &kernel_record->cycle_timestamp,
			  &kernel_record->task_ID);
    break;

  case TAG_FORK:
    items_to_parse = 3;
    parse_result = sscanf(&buffer[2],
			  "%qx %hd %hd",
			  &kernel_record->cycle_timestamp,
			  &kernel_record->task_ID,
			  &kernel_record->parent_task_ID);
    break;

  case TAG_EXEC:
    items_to_parse = 6;
    parse_result = sscanf(&buffer[2],
			  "%qx %hd %lx %hd %hd %s",
			  &kernel_record->cycle_timestamp,
			  &kernel_record->task_ID,
			  &kernel_record->file_ID.inode_ID,
			  &kernel_record->file_ID.major,
			  &kernel_record->file_ID.minor,
			  kernel_record->filename);
    break;

  case TAG_CONTEXT_ASSIGNMENT:
    items_to_parse = 3;
    parse_result = sscanf(&buffer[2],
			  "%qx %hd %lx",
			  &kernel_record->cycle_timestamp,
			  &kernel_record->task_ID,
			  &kernel_record->context_ID);
    break;

  case TAG_DUPLICATE_RANGE:
    items_to_parse = 5;
    parse_result = sscanf(&buffer[2],
			  "%qx %lx %lx %lx %lx",
			  &kernel_record->cycle_timestamp,
			  &kernel_record->context_ID,
			  &kernel_record->duplicate_context_ID,
			  &starting_virtual_address,
			  &ending_virtual_address);
    break;

  case TAG_MMAP_FILE:
    items_to_parse = 9;
    parse_result = sscanf(&buffer[2],
			  "%qx %hd %lx %lx %lx %hd %hd %qx %s",
			  &kernel_record->cycle_timestamp,
			  &kernel_record->task_ID,
			  &starting_virtual_address,
			  &length,
			  &kernel_record->file_ID.inode_ID,
			  &kernel_record->file_ID.major,
			  &kernel_record->file_ID.minor,
			  &file_offset,
			  kernel_record->filename);
    break;

  case TAG_MMAP_ANONYMOUS:
    items_to_parse = 4;
    parse_result = sscanf(&buffer[2],
			  "%qx %hd %lx %lx",
			  &kernel_record->cycle_timestamp,
			  &kernel_record->task_ID,
			  &starting_virtual_address,
			  &length);
    break;

  case TAG_MUNMAP:
    items_to_parse = 4;
    parse_result = sscanf(&buffer[2],
			  "%qx %hd %lx %lx",
			  &kernel_record->cycle_timestamp,
			  &kernel_record->task_ID,
			  &starting_virtual_address,
			  &length);
    break;

  case TAG_COMPLETE_UNMAP:
    items_to_parse = 2;
    parse_result = sscanf(&buffer[2],
			  "%qx %lx",
			  &kernel_record->cycle_timestamp,
			  &kernel_record->context_ID);
    break;

  case TAG_IPC_SHMAT:
    items_to_parse = 5;
    parse_result = sscanf(&buffer[2],
			  "%qx %hd %lx %lx %lx",
			  &kernel_record->cycle_timestamp,
			  &kernel_record->task_ID,
			  &starting_virtual_address,
			  &length,
			  &kernel_record->shm_ID);
    break;

  case TAG_IPC_SHMDT:
    items_to_parse = 3;
    parse_result = sscanf(&buffer[2],
			  "%qx %hd %lx",
			  &kernel_record->cycle_timestamp,
			  &kernel_record->task_ID,
			  &starting_virtual_address);
    break;

  case TAG_COW_UNMAP:
    items_to_parse = 3;
    parse_result = sscanf(&buffer[2],
			  "%qx %hd %lx",
			  &kernel_record->cycle_timestamp,
			  &kernel_record->task_ID,
			  &starting_virtual_address);
    break;

  case TAG_BUFFER_ALLOCATION:
  case TAG_BUFFER_DEALLOCATION:
  case TAG_FILE_CACHE_DEALLOCATION:
    items_to_parse = 2;
    parse_result = sscanf(&buffer[2],
			  "%qx %lx",
			  &kernel_record->cycle_timestamp,
			  &starting_virtual_address);
    break;

  case TAG_FILE_CACHE_ALLOCATION:
    items_to_parse = 6;
    parse_result = sscanf(&buffer[2],
			  "%qx %lx %lx %hd %hd %qx",
			  &kernel_record->cycle_timestamp,
			  &starting_virtual_address,
			  &kernel_record->file_ID.inode_ID,
			  &kernel_record->file_ID.major,
			  &kernel_record->file_ID.minor,
			  &file_offset);
    break;

  case TAG_FILE_OPEN:
    items_to_parse = 6;
    parse_result = sscanf(&buffer[2],
			  "%qx %hd %lx %hd %hd %s",
			  &kernel_record->cycle_timestamp,
			  &kernel_record->task_ID,
			  &kernel_record->file_ID.inode_ID,
			  &kernel_record->file_ID.major,
			  &kernel_record->file_ID.minor,
			  kernel_record->filename);
    break;

  case TAG_FILE_CLOSE:
    items_to_parse = 5;
    parse_result = sscanf(&buffer[2],
			  "%qx %hd %lx %hd %hd",
			  &kernel_record->cycle_timestamp,
			  &kernel_record->task_ID,
			  &kernel_record->file_ID.inode_ID,
			  &kernel_record->file_ID.major,
			  &kernel_record->file_ID.minor);
    break;

  case TAG_FILE_READ:
  case TAG_FILE_WRITE:
    items_to_parse = 7;
    parse_result = sscanf(&buffer[2],
			  "%qx %hd %lx %hd %hd %qx %lx",
			  &kernel_record->cycle_timestamp,
			  &kernel_record->task_ID,
			  &kernel_record->file_ID.inode_ID,
			  &kernel_record->file_ID.major,
			  &kernel_record->file_ID.minor,
			  &file_offset,
			  &length);
    break;

  case TAG_FILE_DELETE:
    items_to_parse = 4;
    parse_result = sscanf(&buffer[2],
			  "%qx %lx %hd %hd",
			  &kernel_record->cycle_timestamp,
			  &kernel_record->file_ID.inode_ID,
			  &kernel_record->file_ID.major,
			  &kernel_record->file_ID.minor);
    break;

  case TAG_END_OF_TRACE:

    fprintf(stderr, "Consumer::read_kernel_record(): ");
    fprintf(stderr, "Premature end of kernel trace\n");
    exit(1);
    break;

  default:
    fprintf(stderr, "Consumer::read_kernel_record(): ");
    fprintf(stderr,
	    "Invalid kernel tag %c(%d)\n",
	    kernel_record->tag,
	    (int)kernel_record->tag);
    exit(1);

  }

  // If the parse failed, emit the error and exit.
  if (parse_result != items_to_parse) {
    fprintf(stderr, "Consumer::read_kernel_record(): ");
    fprintf(stderr, "Could not parse remainder\n");
    fprintf(stderr,
	    "cycle_timestamp = %qx\n",
	    kernel_record->cycle_timestamp);
    exit(1);
  }

  // If the record type requires it, convert the given information
  // about the mapped region into starting and ending virtual page
  // numbers.  Note that some records only use the starting virtual
  // address, but no harm is done by computing a bogus ending address.
  if ((kernel_record->tag == TAG_MMAP_FILE) ||
      (kernel_record->tag == TAG_MMAP_ANONYMOUS) ||
      (kernel_record->tag == TAG_IPC_SHMAT) ||
      (kernel_record->tag == TAG_IPC_SHMDT) ||
      (kernel_record->tag == TAG_MUNMAP) ||
      (kernel_record->tag == TAG_COW_UNMAP) ||
      (kernel_record->tag == TAG_DUPLICATE_RANGE) ||
      (kernel_record->tag == TAG_BUFFER_ALLOCATION) ||
      (kernel_record->tag == TAG_BUFFER_DEALLOCATION) ||
      (kernel_record->tag == TAG_FILE_CACHE_ALLOCATION) ||
      (kernel_record->tag == TAG_FILE_CACHE_DEALLOCATION)) {

    // Subtract one because the ending address is always the first
    // byte *beyond* the given range.
    if (kernel_record->tag == TAG_DUPLICATE_RANGE) {
      ending_virtual_address--;
    } else {
      ending_virtual_address = starting_virtual_address + length - 1;
    }

    kernel_record->starting_page =
      starting_virtual_address >> PAGE_SIZE_ORDER;
    kernel_record->ending_page =
      ending_virtual_address >> PAGE_SIZE_ORDER;

    // For some records, the file offset and length must be converted
    // into starting and ending pages within that address space as
    // well.
    if ((kernel_record->tag == TAG_MMAP_FILE) ||
	(kernel_record->tag == TAG_FILE_CACHE_ALLOCATION) ||
	(kernel_record->tag == TAG_FILE_READ) ||
	(kernel_record->tag == TAG_FILE_WRITE)) {

      kernel_record->file_starting_page =
	file_offset >> PAGE_SIZE_ORDER;
      kernel_record->file_ending_page =
	(file_offset + length - 1) >> PAGE_SIZE_ORDER;

    }
  }

  number_kernel_records++;

} // read_kernel_record
// ===================================================================



// ===================================================================
canonical_page_s*
Consumer::get_new_canonical_page (const page_type_t type) {

  // Ensure that we haven't exhausted the canonical ID space.  Since the
  // upper 1 GB of pages is reserved for the shared kernel space, we
  // must stop allocating new canonical pages once we reach that range.
  if (next_canonical_page_ID == KERNEL_CANONICAL_PAGE_OFFSET) {
    fprintf(stderr,
	    "get_new_canonical_page(): Canonical page IDs exhausted\n");
    exit(1);
  }

  // Allocate a new canonical page structure.
  canonical_page_s* page = new canonical_page_s;
  if (page == 0) {
    fprintf(stderr, "get_new_canonical_page(): Allocation failed\n");
    exit(1);
  }

  // Assign the fields and return a pointer to it.  Assume initially
  // that there are no maps pointing to this page.
  page->page = next_canonical_page_ID++;
  page->reference_count = 0;
  page->type = type;
  return page;

} // get_new_canonical_page
// ===================================================================



// ===================================================================
// Take a handle to the canonical page, making it possible to nullify
// the pointer to the page if it is deleted by this method.

void
Consumer::delete_if_unused_page
(canonical_page_s** const canonical_page_handle) {

  // Is this an anonymous page that is no longer mapped by any
  // context?
  if (((*canonical_page_handle)->type == TYPE_ANONYMOUS) &&
      ((*canonical_page_handle)->reference_count == 0)) {

    // Yes.  Delete this canonical page, as it can never be mapped
    // again.
    delete (*canonical_page_handle);
    (*canonical_page_handle) = 0;

  }
} // Consumer::delete_if_unused_page
// ===================================================================



// ===================================================================
// Take a handle to the process, making it possible to nullify the
// pointer to the process if it is deleted by this method.

void
Consumer::delete_if_defunct_process (Process** const process_handle) {

  Process* process = *process_handle;

  // Are there any tasks left for this process?
  if (process->task_list.empty()) {

    // None.  Unmap any remaining pages in the context.
    handle_complete_unmap(process->context_ID);

    // Find and erase the context_ID->process mapping.
    context_ID_to_Process_map_t::iterator i =
      context_ID_to_Process_map.find(process->context_ID);
    if (i != context_ID_to_Process_map.end()) {

      context_ID_to_Process_map.erase(i);

    }

    // Delete the virtual->canonical map and the process object.
    delete process->virtual_to_canonical_map;
    delete process;
    *process_handle = 0;

  } // Process has no remaining tasks

} // Consumer::delete_if_defunct_process
// ===================================================================



// ===================================================================
void
Consumer::delete_defunct_shm_segments () {

  // Traverse the table of defunct IPC shared memory segments.
  defunct_shm_ID_map_t::iterator i = defunct_shm_ID_map.begin();
  while (i != defunct_shm_ID_map.end()) {

    // Each entry should be marked as ``true'' to indicate that the
    // shm segment is defunct.
    assert((*i).second);

    // Grab the segment ID and lookup its virtual->canonical map.
    shm_ID_t shm_ID = (*i).first;
    shm_ID_to_V2C_map_t::iterator map_iterator =
      shm_ID_to_V2C_map.find(shm_ID);
    assert(map_iterator != shm_ID_to_V2C_map.end());
    virtual_to_canonical_map_t* map = (*map_iterator).second;

    // Traverse the map of canonical pages for this defunct segment.
    virtual_to_canonical_map_t::iterator j = (*map).begin();
    while (j != (*map).end()) {

      // Sanity check the canonical page and then delete it.
      canonical_page_s* canonical_page = (*j).second;
      assert(canonical_page->type == TYPE_IPC_SHM_SEGMENT);
      assert(canonical_page->reference_count == 0);
      delete canonical_page;
      canonical_page = 0;

      // Advance to the next page.  Don't bother erasing the map
      // entry, as we will later delete the entire map.
      j++;

    } // Canonical page traversal

    // All of the canonical pages are gone.  Eliminate the
    // virtual->canonical map itself.
    delete map;
    map = 0;
    shm_ID_to_V2C_map.erase(map_iterator);

    // Delete the entry in the defunct segment table and move on to
    // the next one.
    defunct_shm_ID_map_t::iterator moribund = i;
    i++;
    defunct_shm_ID_map.erase(moribund);

  } // Defunct shm segment traversal

} // Consumer::delete_defunct_shm_segments
// ===================================================================



// ===================================================================
void
Consumer::mark_if_defunct_shm_segment
(const canonical_page_s* const canonical_page) {

  // Is this a canonical page from an IPC shared memory segment that
  // is unattached to any virtual address space
  assert(canonical_page->type == TYPE_IPC_SHM_SEGMENT);
  if (canonical_page->reference_count == 0) {

    // It is.  Mark the segment to which this canonical page belongs
    // as defunct.
    defunct_shm_ID_map[canonical_page->shm_ID] = true;

  }
} // Consumer::mark_if_defunct_shm_segment
// ===================================================================



// ===================================================================
void
Consumer::unlink_task_and_process (Task* const task) {

  // Grab a pointer the task's process.
  Process* process = task->process;

  // If the task is not associated with a process, then there's no
  // work to do here.
  if (process == NULL) {
    return;
  }

  // Clear the process pointer in the task object, thus performing
  // the first 1/2 of the unlinking.
  task->process = NULL;

  // Search through the process's task list for this task.
  list<Task*>::iterator i = process->task_list.begin();
  bool done = false;
  while (!done) {

    // Sanity check:  We should never reach the end of the list
    // without finding the task.
    if (i == process->task_list.end()) {
      fprintf(stderr, "Consumer::unlink_task_and_process(): ");
      fprintf(stderr, "Task not found in process\n");
      exit(1);
    }

    // If we've found it, delete it and end the search.  Otherwise,
    // keep searching.
    if (*i == task) {

      // It is.  Erase it from the list, thus performing the second
      // 1/2 of the unlinking.
      process->task_list.erase(i);
      done = true;

    } else {

      // It is not.  Continue searching.
      i++;

    }
  } // Loop over task IDs in the process

  // The process may be defunct as a result of the unlinking, and if
  // so, it should be deleted.
  delete_if_defunct_process(&process);

} // Consumer::unlink_task_and_process
// ===================================================================



// ===================================================================
void
Consumer::make_kernel_active () {

  // It must be the case that the scheduled task was previously
  // active.
  if (active_task != scheduled_task) {
    fprintf(stderr, "Consumer::make_kernel_active(): ");
    fprintf(stderr, "User task was not active\n");
    exit(1);
  }

  // Mark the beginning of a quanta (of sorts) for the kernel.
  kernel_task->
    mark_quanta_start(reference_record->cycle_timestamp,
		      reference_record->instruction_timestamp,
		      reference_record->reference_timestamp);

  // Make the kernel the active task.
  active_task = kernel_task;

} // Consumer::make_kernel_active
// ===================================================================



// ===================================================================
void
Consumer::make_user_active () {

  // It must be the case that the kernel task was previously active.
  if (active_task != kernel_task) {
    fprintf(stderr, "Consumer::make_kernel_active(): ");
    fprintf(stderr, "Kernel task was not active\n");
    exit(1);
  }
  // Mark the beginning of a quanta (of sorts) for the user.
  scheduled_task->
    mark_quanta_start(reference_record->cycle_timestamp,
		      reference_record->instruction_timestamp,
		      reference_record->reference_timestamp);

  // Make the user the active task.
  active_task = scheduled_task;

} // Consumer::make_user_active
// ===================================================================



// ===================================================================
void
Consumer::handle_schedule () {

  // Find the task object based on the given ID.
  Task* newly_scheduled_task =
    safe_task_ID_to_task(kernel_record->task_ID);

  // Is the currently scheduled task the active one?
  if (active_task == scheduled_task) {

    // Also mark the starting time of the new task's quanta, thus
    // allowing it to maintain its virtual time.
    newly_scheduled_task->
      mark_quanta_start(reference_record->cycle_timestamp,
			reference_record->instruction_timestamp,
			reference_record->reference_timestamp);

  }

  // Assign the newly scheduled task.
  scheduled_task = newly_scheduled_task;

} // Consumer::handle_schedule
// ===================================================================



// ===================================================================
void
Consumer::handle_fork () {

  // Sanity check: There should be no task already using the given ID.
  task_ID_to_Task_map_t::iterator i =
    task_ID_to_Task_map.find(kernel_record->task_ID);
  if (i != task_ID_to_Task_map.end()) {
    fprintf(stderr, "Consumer::handle_fork(): Duplicate task ID\n");
    exit(1);
  }

  // Create a new task object.
  Task* child_task = allocate_new_task();

  // Look up the parent task.
  Task* parent_task =
    safe_task_ID_to_task(kernel_record->parent_task_ID);

  // Assign the new task's fields.  Note that until the task is
  // assigned a context, it won't be associated with a process.
  child_task->task_ID = kernel_record->task_ID;
  child_task->process = NULL;
  child_task->parent_task = parent_task;

  // Create a table entry mapping the task ID to the new task object.
  task_ID_to_Task_map[kernel_record->task_ID] = child_task;

} // Consumer::handle_fork
// ===================================================================



// ===================================================================
void
Consumer::handle_exec () {

  // Grab the task process associated with this task.
  Task* task = safe_task_ID_to_task(kernel_record->task_ID);
  Process* process = task->process;

  // Assign the new executable image.
  process->image_file_ID = kernel_record->file_ID;
  strcpy(process->image_pathname, kernel_record->filename);

} // Consumer::handle_exec
// ===================================================================



// ===================================================================
void
Consumer::handle_exit () {

  // Lookup the task and process objects for the given task ID.
  task_ID_to_Task_map_t::iterator task_iterator =
    task_ID_to_Task_map.find(kernel_record->task_ID);
  if (task_iterator == task_ID_to_Task_map.end()) {
    fprintf(stderr, "Consumer::handle_exit(): Failed lookup\n");
    exit(1);
  }
  Task* task = (*task_iterator).second;

  // Separate the task and its process.  If the process becomes
  // defunct as a result, it will be deleted.
  unlink_task_and_process(task);

  // Clear the task_ID->process mapping.
  task_ID_to_Task_map.erase(task_iterator);

  // Delete the task itself.
  delete task;
  task = 0;

} // Consumer::handle_exit
// ===================================================================



// ===================================================================
void
Consumer::handle_context_assignment () {

  // Grab the task object for the given task ID.
  Task* task = safe_task_ID_to_task(kernel_record->task_ID);

  // Unlink this task from its current process (if any).
  unlink_task_and_process(task);

  // Get the mapping from the given context ID to its process.  Is it
  // mapped to any process?
  context_ID_to_Process_map_t:: iterator process_iterator =
    context_ID_to_Process_map.find(kernel_record->context_ID);
  Process* process;
  if (process_iterator == context_ID_to_Process_map.end()) {

    // The context ID is not yet associated with a process.  Allocate
    // a new process object.
    process = new Process;
    if (process == 0) {
      fprintf(stderr, "Consumer::handle_context_assignment(): ");
      fprintf(stderr, "Failed allocation of new process\n");
      exit(1);
    }

    // Assign the process a new ID.
    process->process_ID = next_process_ID++;
    if (process->process_ID == 0) {
      fprintf(stderr, "Consumer::handle_context_assignment(): ");
      fprintf(stderr, "Exhausted process ID space\n");
      exit(1);
    }

    // Create a new context by creating a new virtual->canonical map,
    // and assign it to the new process.
    process->virtual_to_canonical_map = new virtual_to_canonical_map_t;
    if (process->virtual_to_canonical_map == NULL) {
      fprintf(stderr, "Consumer::handle_context_assignment(): ");
      fprintf(stderr, "Failed allocation of virtual->canonical map\n");
      exit(1);
    }

    // Assign the given context ID to the process.
    process->context_ID = kernel_record->context_ID;

    // Assign a mapping from this context ID to this process.
    context_ID_to_Process_map[process->context_ID] = process;

    // Assign the process an executable image by copying it from this
    // task's parent's process.
    Process* parent_process = task->parent_task->process;
    process->image_file_ID = parent_process->image_file_ID;
    strcpy(process->image_pathname, parent_process->image_pathname);

  } else {

    // A process already exists using the given context, so this task
    // is joining an existing process.  Grab that process.
    process = (*process_iterator).second;

    // Sanity check:  This process should explicitly indicate that it
    // is using the given context ID.
    if (process->context_ID != kernel_record->context_ID) {
      fprintf(stderr, "Consumer::handle_context_assignment(): ");
      fprintf(stderr, "Process and map disagree on context ID\n");
      exit(1);
    }

  }

  // Link the task and the process to one another.
  task->process = process;
  process->task_list.push_front(task);

} // Consumer::handle_context_assignment
// ===================================================================



// ===================================================================
void
Consumer::handle_duplicate_range () {

  // Get the virtual->canonical map for both the original (parent) and
  // duplicate (child) contexts.
  Process* original_process =
    safe_context_ID_to_process(kernel_record->context_ID);
  virtual_to_canonical_map_t* original_map =
    original_process->virtual_to_canonical_map;
  Process* duplicate_process =
    safe_context_ID_to_process(kernel_record->duplicate_context_ID);
  virtual_to_canonical_map_t* duplicate_map =
    duplicate_process->virtual_to_canonical_map;

  // Traverse every virtual page in the range, making that page in the
  // duplicate context point to the same canonical page as that page
  // does in the original context.
  for (virtual_page_ID_t current_page = kernel_record->starting_page;
       current_page <= kernel_record->ending_page;
       current_page++) {

    // We should never be trying to duplicate pages in the kernel's
    // address range.
    if (current_page >= KERNEL_VIRTUAL_PAGE_OFFSET) {
      fprintf(stderr, "Consumer::handle_duplicate_range(): ");
      fprintf(stderr, "In kernel range\n");
      exit(1);
    }

    // Attempt to lookup the canonical page underlying the original.
    // Note that it is possible for a task to map an anonymous
    // region without having referenced it yet, leaving some pages in
    // that region without assigned canonical pages.
    virtual_to_canonical_map_t::iterator i =
      (*original_map).find(current_page);

    // Has a canonical page yet been assigned to this page in the
    // original context?
    canonical_page_s* canonical_page = 0;
    if (i == (*original_map).end()) {

      // No.  Create a new canonical page and map it into the original
      // context.
      canonical_page = get_new_canonical_page(TYPE_ANONYMOUS);
      map_virtual_to_canonical(original_map,
			     current_page,
			     canonical_page);
      canonical_page->reference_count++;

    } else {

      // Yes.  Use that canonical page.
      canonical_page = (*i).second;

    }

    // The current page in the duplicate context cannot be mapped to
    // anything -- an error has occurred if it is.
    if ((*duplicate_map).find(current_page) !=
	(*duplicate_map).end()) {
      fprintf(stderr, "Consumer::handle_duplicate_range(): ");
      fprintf(stderr, "Duplicate range not unmapped\n");
      exit(1);
    }

    // Map this canonical page into the duplicate context.
    map_virtual_to_canonical(duplicate_map,
			     current_page,
			     canonical_page);
    canonical_page->reference_count++;

  }
} // Consumer::handle_duplicate_range
// ===================================================================



// ===================================================================
void
Consumer::handle_file_mmap_range () {

  // Get the virtual->canonical maps for the context and the file.
  Task* task = safe_task_ID_to_task(kernel_record->task_ID);
  virtual_to_canonical_map_t* context_map =
    task->process->virtual_to_canonical_map;
  virtual_to_canonical_map_t* file_map =
    safe_file_ID_to_V2C_map(kernel_record->file_ID);

  // Traverse each page in the range, establishing an entry that maps
  // the task virtual page to the file page.
  virtual_page_ID_t task_page = kernel_record->starting_page;
  virtual_page_ID_t file_page = kernel_record->file_starting_page;
  while (task_page <= kernel_record->ending_page) {

    // The task page should never be in the kernel's page range.
    if (task_page >= KERNEL_VIRTUAL_PAGE_OFFSET) {
      fprintf(stderr, "Consumer::handle_file_mmap_range(): ");
      fprintf(stderr, "In kernel range\n");
      exit(1);
    }

    // Attempt to lookup the canonical page underlying this file page.
    virtual_to_canonical_map_t::iterator i =
      (*file_map).find(file_page);

    // Has a canonical page yet been assigned to this file page?
    canonical_page_s* canonical_page = 0;
    if (i == (*file_map).end()) {

      // No.  Create a new canonical page and map it into the file's
      // space.
      canonical_page = get_new_canonical_page(TYPE_FILE_CACHE);
      map_virtual_to_canonical(file_map, file_page, canonical_page);
      
    } else {

      // Yes.  Use that canonical page.
      canonical_page = (*i).second;

      // Is this page not an mmap page?
      if (canonical_page->type != TYPE_FILE_CACHE) {

	// This is not a file cache page, so something has gone wrong
	// in mapping a canonical page that was of some other type.
	fprintf(stderr, "Consumer::handle_file_mmap_range(): ");
	fprintf(stderr, "Mapping canonical page of wrong type\n");
	exit(1);

      }
    }

    // Ensure that the task doesn't already have a mapping for this
    // page in its space.  If so, emit the error and exit.
    if ((*context_map).find(task_page) != (*context_map).end()) {
      fprintf(stderr, "Consumer::handle_file_mmap_range(): ");
      fprintf(stderr, "Attempt to map an already mapped page\n");
      exit(1);
    }

    // Map this canonical page into the task's space.
    map_virtual_to_canonical(context_map, task_page, canonical_page);
    canonical_page->reference_count++;
 
    // Move to the next page in each space.
    task_page++;
    file_page++;

  }

} // Consumer::handle_file_mmap_range
// ===================================================================



// ===================================================================
void
Consumer::handle_anonymous_mmap_range () {

  // Grab the virtual->canonical map for the process's context.
  Task* task = safe_task_ID_to_task(kernel_record->task_ID);
  virtual_to_canonical_map_t* map =
    task->process->virtual_to_canonical_map;

  // Traverse each page in the range, establishing an entry that maps
  // the task virtual page to the canonical ``zero page''.
  virtual_page_ID_t page = kernel_record->starting_page;
  while (page <= kernel_record->ending_page) {

    // The task page should never be in the kernel's page range.
    if (page >= KERNEL_VIRTUAL_PAGE_OFFSET) {
      fprintf(stderr, "Consumer::handle_anonymous_mmap_range(): ");
      fprintf(stderr, "In kernel range\n");
      exit(1);
    }

    // Ensure that the task doesn't already have a mapping for this
    // page in its space.  If so, emit the error and exit.
    if ((*map).find(page) != (*map).end()) {
      fprintf(stderr, "Consumer::handle_anonymous_mmap_range(): ");
      fprintf(stderr, "Attempt to map an already mapped page\n");
      exit(1);
    }

    // Map the canonical zero page into the task's space.
    map_virtual_to_canonical(map, page, canonical_zero_page);
    canonical_zero_page->reference_count++;
 
    // Move to the next page.
    page++;

  }
} // Consumer::handle_anonymous_mmap_range
// ===================================================================



// ===================================================================
void
Consumer::handle_shmat () {

  // Grab the virtual->canonical maps for the context and the shared
  // memory segment.
  Task* task = safe_task_ID_to_task(kernel_record->task_ID);
  virtual_to_canonical_map_t* context_map =
    task->process->virtual_to_canonical_map;
  virtual_to_canonical_map_t* shm_map =
    safe_shm_ID_to_V2C_map(kernel_record->shm_ID);

  // Traverse every page in the range, establishing an entry that maps
  // the task virtual page to the shared memory segment page.  Note
  // that shared memory segments are assumed to start at an offset of
  // zero.
  virtual_page_ID_t task_page = kernel_record->starting_page;
  virtual_page_ID_t shm_page = 0;
  while (task_page <= kernel_record->ending_page) {

    // The task page should never be in the kernel's page range.
    if (task_page >= KERNEL_VIRTUAL_PAGE_OFFSET) {
      fprintf(stderr, "Consumer::handle_shmat(): In kernel range\n");
      exit(1);
    }

    // Attempt to lookup the canonical page underlying this shared
    // memory page.
    virtual_to_canonical_map_t::iterator i = (*shm_map).find(shm_page);

    // Has a canonical page yet been assigned to this shared memory
    // page?
    canonical_page_s* canonical_page = 0;
    if (i == (*shm_map).end()) {

      // No.  Create a new canonical page, associate it with this new
      // shm segment, and map it into the shared memory segment's
      // space.
      canonical_page = get_new_canonical_page(TYPE_IPC_SHM_SEGMENT);
      canonical_page->shm_ID = kernel_record->shm_ID;
      map_virtual_to_canonical(shm_map, shm_page, canonical_page);
      
    } else {

      // Yes.  Use that canonical page.
      canonical_page = (*i).second;

    }

    // Ensure that the task doesn't already have a mapping for this
    // page in its space.  If so, emit the error and exit.
    if ((*context_map).find(task_page) != (*context_map).end()) {
      fprintf(stderr, "Consumer::handle_shmat(): ");
      fprintf(stderr, "Attempt to map an already mapped page\n");
      exit(1);
    }

    // Map this canonical page into the task's space.
    map_virtual_to_canonical(context_map, task_page, canonical_page);
    canonical_page->reference_count++;

    // Move to the next pair of pages.
    task_page++;
    shm_page++;

  }
} // Consumer::handle_shmat
// ===================================================================



// ===================================================================
void
Consumer::handle_munmap_range () {

  // Grab the virtual->canonical map for the context.
  Task* task = safe_task_ID_to_task(kernel_record->task_ID);
  virtual_to_canonical_map_t* map =
    task->process->virtual_to_canonical_map;

  // Traverse every page in the range, eliminating any mappings
  // found.
  virtual_page_ID_t page = kernel_record->starting_page;
  while (page <= kernel_record->ending_page) {

    // The task page should never be in the kernel's page range.
    if (page >= KERNEL_VIRTUAL_PAGE_OFFSET) {
      fprintf(stderr, "Consumer::handle_munmap_range(): ");
      fprintf(stderr, "In kernel range\n");
      exit(1);
    }

    // Is this page mapped to another?
    virtual_to_canonical_map_t::iterator i = (*map).find(page);
    if (i != (*map).end()) {

      // It is, so eliminate the mapping.  If the underlying canonical
      // pages indicate that this is an IPC shared memory segment,
      // determine if that segment is defunct and should be deleted.
      canonical_page_s* canonical_page = (*i).second;
      canonical_page->reference_count--;
      (*map).erase(i);
      delete_if_unused_page(&canonical_page);

      // Note that only anonymous pages are deleted by the above call,
      // so this use of canonical_page is safe.
      if ((canonical_page != NULL) &&
	  (canonical_page->type == TYPE_IPC_SHM_SEGMENT)) {
	mark_if_defunct_shm_segment(canonical_page);
      }

    }

    // Move to the next page.
    page++;

  }
} // Consumer::handle_munmap_range
// ===================================================================



// ===================================================================
void
Consumer::handle_complete_unmap (const context_ID_t context_ID) {

  // Is this context ID mapped to a process?  It may not be if the
  // process with which it was associated became defunct and was
  // deleted, taking the context with it.
  context_ID_to_Process_map_t::iterator process_iterator =
    context_ID_to_Process_map.find(context_ID);
  if (process_iterator == context_ID_to_Process_map.end()) {

    // The context has been deleted.  No work to do here.
    return;

  }

  // Grab the virtual->canonical map that forms the context for the
  // process.
  virtual_to_canonical_map_t* map =
    (*process_iterator).second->virtual_to_canonical_map;

  // This function takes a brute-force approach: It traverses every
  // page in the table, eliminating any mappings it might find.
  virtual_to_canonical_map_t::iterator i = (*map).begin();
  while (i != (*map).end()) {

    // It should never be the case that we are unmapping pages in the
    // kernel's page range.
    if ((*i).first >= KERNEL_VIRTUAL_PAGE_OFFSET) {
      fprintf(stderr, "Consumer::handle_complete_unmap(): ");
      fprintf(stderr, "In kernel range\n");
      exit(1);
    }

    // Eliminate the mapping.  If the underlying canonical pages
    // indicate that this is an IPC shared memory segment, determine
    // if that segment is defunct and should be deleted.
    canonical_page_s* canonical_page = (*i).second;
    virtual_to_canonical_map_t::iterator moribund = i;
    i++;
    (*map).erase(moribund);
    canonical_page->reference_count--;
    delete_if_unused_page(&canonical_page);

    // Note that only anonymous pages are deleted by the above call,
    // so if this is an IPC shm canonical page, it will still exist.
    if ((canonical_page != 0) &&
	(canonical_page->type == TYPE_IPC_SHM_SEGMENT)) {
      mark_if_defunct_shm_segment(canonical_page);
    }

  }
} // Consumer::handle_complete_unmap
// ===================================================================



// ===================================================================
void
Consumer::handle_cow_unmap () {

  virtual_page_ID_t page = kernel_record->starting_page;

  // The page should not be in the kernel's page range.
  if (page >= KERNEL_VIRTUAL_PAGE_OFFSET) {
    fprintf(stderr, "Consumer::handle_cow_unmap(): In kernel range\n");
    exit(1);
  }

  // Grab the virtual->canonical map for the context.
  Task* task = safe_task_ID_to_task(kernel_record->task_ID);
  virtual_to_canonical_map_t* map =
    task->process->virtual_to_canonical_map;

  // This page should be mapped.  If it is not, an error has
  // occurred.
  virtual_to_canonical_map_t::iterator i = (*map).find(page);
  if (i == (*map).end()) {
    fprintf(stderr, "Consumer::handle_cow_unmap(): ");
    fprintf(stderr, "Unmapping a page with no mapping\n");
    exit(1);
  }

  // Reassign this virtual page to a new canonical page, thus
  // representing the copy that is made by the COW mechanism.  Whether
  // the original page was anonymous or from a file, the new copy must
  // be anonymous.  Also increase its sharing count.
  canonical_page_s* originalCanonicalPage = (*i).second;
  originalCanonicalPage->reference_count--;
  (*map).erase(i);
  canonical_page_s* newCanonicalPage =
    get_new_canonical_page(originalCanonicalPage->type);
  newCanonicalPage->type = TYPE_ANONYMOUS;
  newCanonicalPage->reference_count++;
  map_virtual_to_canonical(map, page, newCanonicalPage);

  // The original page cannot have been an IPC shared memory page.
  if (originalCanonicalPage->type == TYPE_IPC_SHM_SEGMENT) {
    fprintf(stderr, "Consumer::handle_cow_unmap(): ");
    fprintf(stderr, "COW with an IPC shm page\n");
    exit(1);
  }

  // When COW-unmapping an anonymous page, it should never be the case
  // that the sharing count reaches zero -- the last task to write
  // to the COW page should get that original.
  if ((originalCanonicalPage->type == TYPE_ANONYMOUS) &&
      (originalCanonicalPage->reference_count == 0)) {
    fprintf(stderr, "Consumer::handle_cow_unmap(): ");
    fprintf(stderr, "Anonymous page orphaned\n");
    exit(1);
  }
} // Consumer::handle_cow_unmap
// ===================================================================



// ===================================================================
void
Consumer::handle_buffer_cache_allocation () {

  virtual_page_ID_t page = kernel_record->starting_page;

  // The page must be in the kernel's page range.
  if (page < KERNEL_VIRTUAL_PAGE_OFFSET) {
    fprintf(stderr, "Consumer::handle_buffer_cache_allocation(): ");
    fprintf(stderr, "Outside kernel range\n");
    exit(1);
  }

  // Grab the virtual->canonical map for the kernel's context.
  Process* kernel_process =
    safe_context_ID_to_process(KERNEL_CONTEXT);
  virtual_to_canonical_map_t* map =
    kernel_process->virtual_to_canonical_map;

  // This page should not be mapped.
  virtual_to_canonical_map_t::iterator i = (*map).find(page);
  if (i != (*map).end()) {
    fprintf(stderr, "Consumer::handle_buffer_cache_allocation(): ");
    fprintf(stderr,
	    "At record %qd: Page already mapped\n",
	    number_kernel_records);
    exit(1);
  }

  // Assign this virtual page to the one canonical buffer cache page.
  // Increment the reference count so that we can track the number of
  // buffer cache pages at each moment.
  canonical_buffer_cache_page.reference_count++;
  map_virtual_to_canonical(map, page, &canonical_buffer_cache_page);

} // Consumer::handle_buffer_cache_allocation
// ===================================================================



// ===================================================================
void
Consumer::handle_buffer_cache_deallocation () {

  virtual_page_ID_t page = kernel_record->starting_page;

  // The page must be in the kernel's page range.
  if (page < KERNEL_VIRTUAL_PAGE_OFFSET) {
    fprintf(stderr, "Consumer::handle_buffer_cache_deallocation(): ");
    fprintf(stderr, "Outside kernel range\n");
    exit(1);
  }

  // Grab the virtual->canonical map for the kernel's context.
  Process* kernel_process =
    safe_context_ID_to_process(KERNEL_CONTEXT);
  virtual_to_canonical_map_t* map =
    kernel_process->virtual_to_canonical_map;

  // This page must be mapped to the canonical buffer cache page.
  virtual_to_canonical_map_t::iterator i = (*map).find(page);
  if (i == (*map).end()) {
    fprintf(stderr, "Consumer::handle_buffer_cache_deallocation(): ");
    fprintf(stderr, "Page not mapped\n");
    exit(1);
  }
  canonical_page_s* canonical_page = (*i).second;
  if (canonical_page->type != TYPE_BUFFER_CACHE) {
    fprintf(stderr, "Consumer::handle_buffer_cache_deallocation(): ");
    fprintf(stderr,
	    "At %qd, deallocating non-buffer cache page\n",
	    number_kernel_records);
    exit(1);
  }	    

  // Eliminate the mapping.  Decrement the reference count to track
  // the number of buffer cache pages.
  (*map).erase(i);
  canonical_page->reference_count--;

} // Consumer::handle_buffer_cache_deallocation
// ===================================================================



// ===================================================================
void
Consumer::handle_file_cache_allocation () {

  virtual_page_ID_t kernel_page = kernel_record->starting_page;
  virtual_page_ID_t file_page = kernel_record->file_starting_page;

  // The page must be in the kernel's page range.
  if (kernel_page < KERNEL_VIRTUAL_PAGE_OFFSET) {
    fprintf(stderr, "Consumer::handle_file_cache_allocation(): ");
    fprintf(stderr, "Outside kernel range\n");
    exit(1);
  }

  // Grab the virtual->canonical map for the kernel's context and the
  // file.
  Process* kernel_process =
    safe_context_ID_to_process(KERNEL_CONTEXT);
  virtual_to_canonical_map_t* context_map =
    kernel_process->virtual_to_canonical_map;
  virtual_to_canonical_map_t* file_map =
    safe_file_ID_to_V2C_map(kernel_record->file_ID);

  // Attempt to lookup the canonical page underlying this file page.
  virtual_to_canonical_map_t::iterator i = (*file_map).find(file_page);

  // Has a canonical page yet been assigned to this file page?
  canonical_page_s* canonical_page = 0;
  if (i == (*file_map).end()) {

    // No.  Create a new canonical page and map it into the file's
    // space.
    canonical_page = get_new_canonical_page(TYPE_FILE_CACHE);
    map_virtual_to_canonical(file_map, file_page, canonical_page);
      
  } else {

    // Yes.  Use that canonical page.
    canonical_page = (*i).second;

    // Ensure that the canonical page is of the correct type.
    if (canonical_page->type != TYPE_FILE_CACHE) {
      fprintf(stderr, "Consumer::handle_file_cache_allocation(): ");
      fprintf(stderr, "Mapping wrong type of canonical page\n");
      exit(1);

    }
  }

  // Ensure that the KERNEL doesn't already have a mapping for this
  // page in its space.  If so, emit the error and exit.
  if ((*context_map).find(kernel_page) != (*context_map).end()) {
    fprintf(stderr, "Consumer::handle_file_cache_allocation(): ");
    fprintf(stderr, "Attempt to map an already mapped page\n");
    exit(1);
  }

  // Map this canonical page into the task's space.  Note that we do
  // not attribute a reference count to this page, as it may not be
  // mapped into any virtual space.  Since it may be used for VFS
  // access, canonical file cache pages are never freed.
  map_virtual_to_canonical(context_map, kernel_page, canonical_page);

} // Consumer::handle_file_cache_allocation
// ===================================================================



// ===================================================================
void
Consumer::handle_file_cache_deallocation () {

  virtual_page_ID_t page = kernel_record->starting_page;

  // The address must be in the kernel's range.
  if (page < KERNEL_VIRTUAL_PAGE_OFFSET) {
    fprintf(stderr, "Consumer::handle_file_cache_deallocation(): ");
    fprintf(stderr, "Outside kernel range\n");
    exit(1);
  }

  // Get the virtual->canonical map for the kernel.
  Process* kernel_process =
    safe_context_ID_to_process(KERNEL_CONTEXT);
  virtual_to_canonical_map_t* map =
    kernel_process->virtual_to_canonical_map;

  // The page not actually have been recorded as a file cache page (as
  // Linux calls for deallocation on pages that may never have reached
  // the point of insertion into the page cache).
  virtual_to_canonical_map_t::iterator i = (*map).find(page);
  if (i != (*map).end()) {

    canonical_page_s* canonical_page = (*i).second;
    if (canonical_page->type != TYPE_FILE_CACHE) {
      fprintf(stderr, "Consumer::handle_file_cache_deallocation(): ");
      fprintf(stderr, "Deallocating non-file cache page\n");
      exit(1);
    }

    // Eliminate the mapping.
    (*map).erase(i);

  }
} // Consumer::handle_file_cache_deallocation
// ===================================================================



// ===================================================================
void
Consumer::handle_file_delete () {

  // Grab the context for this file.
  file_ID_to_V2C_map_t::iterator map_iterator =
    file_ID_to_V2C_map.find(kernel_record->file_ID);

  // It may be the case that we've not encountered this file before,
  // as a file need not be mapped or opened to be deleted by some
  // process.  There's only work to do if we've a record of this
  // file.
  if (map_iterator != file_ID_to_V2C_map.end()) {

    virtual_to_canonical_map_t* map =
      safe_file_ID_to_V2C_map(kernel_record->file_ID);

    // Traverse every page in the table, deleting the underlying
    // canonical pages.
    virtual_to_canonical_map_t::iterator i = (*map).begin();
    while (i != (*map).end()) {

      // Sanity check: The canonical page should have a reference
      // count of zero, or else the filesystem was deleting a file
      // that had actively open and mapped pages.
      canonical_page_s* canonical_page = (*i).second;
      assert(canonical_page->reference_count == 0);

      // Delete the page.  There's no need to delete the individual
      // mapping, as the whole map will be deleted later.
      delete canonical_page;
      canonical_page = 0;

      // Move to the next mapped page.
      i++;

    } // Table traversal

    // Delete the map and its association with the given file ID.
    delete map;
    map = 0;
    file_ID_to_V2C_map.erase(map_iterator);

  } // File previously encountered

} // handle_file_delete
// ===================================================================



// ===================================================================
void
Consumer::update_with_reference_record () {

  // What type of reference is this?
  switch (reference_record->tag) {

  case TAG_KERNEL_READ:
  case TAG_KERNEL_WRITE:
  case TAG_KERNEL_INSTRUCTION_FETCH:

    // If the kernel is not the active task, then switch to it.
    if (active_task != kernel_task) {
      make_kernel_active();
    }
    break;

  case TAG_USER_READ:
  case TAG_USER_WRITE:
  case TAG_USER_INSTRUCTION_FETCH:

    // If the user-level task is not the active one, make it so.
    if (active_task != scheduled_task) {
      make_user_active();
    }
    break;

  default:

    fprintf(stderr, "Consumer::update_with_reference_record(): ");
    fprintf(stderr,
	    "Invalid reference tag %c(%hd)\n",
	    reference_record->tag,
	    (unsigned short int)reference_record->tag);
    exit(1);

  }

  // Grab local copies of the reference page's information.
  const context_ID_t context_ID =
    reference_record->virtual_page.context_ID;
  const virtual_page_ID_t page_ID =
    reference_record->virtual_page.page_ID;

  // Sanity check:  The context used should be either the kernel's or
  // the scheduled task's.
  if ((context_ID != scheduled_task->process->context_ID) &&
      (context_ID != KERNEL_CONTEXT)) {
    fprintf(stderr, "Consumer::update_with_reference_record(): ");
    fprintf(stderr, "Unexpected context ID used\n");
    exit(1);
  }

  // An update to the page mappings is required only if this record
  // represents a reference not within the kernel's shared space.
  if (page_ID < KERNEL_VIRTUAL_PAGE_OFFSET) {

    // Grab the map of the context in which this reference was
    // performed.
    Process* process = safe_context_ID_to_process(context_ID);
    virtual_to_canonical_map_t* map = process->virtual_to_canonical_map;

    // Is this virtual page mapped to some canonical page?
    virtual_to_canonical_map_t::iterator i = (*map).find(page_ID);
    if (i == (*map).end()) {

      // No.  Create a new canonical page and map this virtual page to
      // it.
      canonical_page_s* canonical_page =
	get_new_canonical_page(TYPE_ANONYMOUS);
      map_virtual_to_canonical(map, page_ID, canonical_page);
      canonical_page->reference_count++;

    }
  }
} // Consumer::update_with_reference_record
// ===================================================================



// ===================================================================
void
Consumer::update_with_kernel_record () {

  // Update the executable image pathname based on the record
  // type.
  switch (kernel_record->tag) {

  case TAG_SCHEDULE:

    handle_schedule();
    break;

  case TAG_FORK:

    handle_fork();
    break;

  case TAG_EXEC:

    handle_exec();
    break;

  case TAG_EXIT:

    handle_exit();
    break;

  case TAG_CONTEXT_ASSIGNMENT:

    handle_context_assignment();
    break;

  case TAG_DUPLICATE_RANGE:

    handle_duplicate_range();
    break;

  case TAG_MMAP_FILE:

    handle_file_mmap_range();
    break;

  case TAG_MMAP_ANONYMOUS:

    handle_anonymous_mmap_range();
    break;

  case TAG_MUNMAP:

    handle_munmap_range();
    break;

  case TAG_COMPLETE_UNMAP:

    handle_complete_unmap(kernel_record->context_ID);
    break;

  case TAG_IPC_SHMAT:

    handle_shmat();
    break;

  case TAG_IPC_SHMDT:

    // At the moment, we rely on the munmap() call performed by every
    // shmdt() call to unmap the IPC shared memory segment pages and
    // to determine whether a particular segment has become defunct
    // and should be deleted.
    break;

  case TAG_COW_UNMAP:

    handle_cow_unmap();
    break;

  case TAG_BUFFER_ALLOCATION:

    handle_buffer_cache_allocation();
    break;

  case TAG_BUFFER_DEALLOCATION:

    handle_buffer_cache_deallocation();
    break;

  case TAG_FILE_CACHE_ALLOCATION:

    handle_file_cache_allocation();
    break;

  case TAG_FILE_CACHE_DEALLOCATION:

    handle_file_cache_deallocation();
    break;

  case TAG_FILE_OPEN:
  case TAG_FILE_CLOSE:
  case TAG_FILE_READ:
  case TAG_FILE_WRITE:
  case TAG_ACCEPT:

    // At the moment, we do nothing here.  We should eventually have
    // data structures tracking file use in detail so that we can emit
    // filesystem traces and sanity check the kernel trace.
    break;

  case TAG_FILE_DELETE:

    handle_file_delete();
    break;

  default:

    fprintf(stderr, "Consumer::update_with_kernel_record(): ");
    fprintf(stderr,
	    "At record %qd: Unknown tag %c(%d)\n",
	    number_kernel_records,
	    kernel_record->tag,
	    (int)kernel_record->tag);
    exit(1);

  }

  // If any unmapping occurred, there may be IPC shared memory
  // segments that are now defunct and need to be deleted.
  delete_defunct_shm_segments();

} // Consumer::update_with_kernel_record
// ===================================================================



// ===================================================================
void
Consumer::lookup_canonical_page_ID
(bool* return_V2C_mapping_exists,
 canonical_page_ID_t* const return_canonical_page_ID,
 const virtual_to_canonical_map_t* const map,
 const virtual_page_ID_t virtual_page_ID) {

  // Assume initially that if the virtual page is valid, its mapping
  // into canonical space will exist.
  *return_V2C_mapping_exists = true;

  // Attempt to lookup the page in the context's map.
  const virtual_to_canonical_map_t::const_iterator i =
    (*map).find(virtual_page_ID);

  // Was the lookup successful?
  if (i != (*map).end()) {

    // Yes.  Is the canonical page a buffer cache page?
    if ((*i).second->type == TYPE_BUFFER_CACHE) {

      // It is a buffer cache page, and so no mapping into canonical
      // space really exists.
      *return_V2C_mapping_exists = false;

    } else {

      // It is not a buffer cache page, so return the canonical page's
      // ID.
      *return_canonical_page_ID = (*i).second->page;

    }

    // The lookup failed.  Is the virtual address in the kernel's
    // address space?
  } else if (virtual_page_ID >= KERNEL_VIRTUAL_PAGE_OFFSET) {

    // Yes.  The upper 1 GB of the canonical space is reserved to
    // correspond to the upper 1 GB of kernel virtual address
    // space. [Linux specific] Calculate the canonical page ID.
    *return_canonical_page_ID =
      KERNEL_CANONICAL_PAGE_OFFSET + (virtual_page_ID -
				      KERNEL_VIRTUAL_PAGE_OFFSET);

  } else {

    // There was no mapping into canonical space found, and the virtual
    // address is not in the kernel space.  Therefore, this is an
    // invalid virtual address.  Emit the error and exit.
    fprintf(stderr, "Consumer::lookup_canonical_page_ID(): ");
    fprintf(stderr, "Failed lookup\n");
    exit(1);

  }
} // lookup_canonical_page_ID
// ===================================================================



// ===================================================================
Task*
Consumer::allocate_new_task () {

  // Attempt to allocate a new Task.  Check for success.
  Task* task = new Task;
  if (task == NULL) {
    fprintf(stderr, "Consumer::allocate_new_task(): ");
    fprintf(stderr, "Allocation failure\n");
    exit(1);
  }

  return task;

} // Consumer::allocate_new_task
// ===================================================================



// ===================================================================
Process*
Consumer::allocate_new_process () {

  // Attempt to allocate a new Task.  Check for success.
  Process* process = new Process;
  if (process == NULL) {
    fprintf(stderr, "Consumer::allocate_new_process(): ");
    fprintf(stderr, "Allocation failure\n");
    exit(1);
  }

  return process;

} // Consumer::allocate_new_process
// ===================================================================
