/**
 * \file Instruction_Reduced_Reader.cc
 * \author Scott F. H. Kaplan <sfkaplan@cs.amherst.edu>
 * \date January 2003
 * \brief Code for a class that reads and parses instruction-level
 * reduced reference traces.
 *
 * A subclass of Reference_Reader that can read the instruction-level
 * reduced reference traces produced by the
 * <i>instruction-level-reducer</i> reference handler.  The data from
 * each record is then transformed appropriately for use by
 * Laplace-merge.
 */



// ===================================================================
// INCLUDES

#include <math.h> // For pow()
#include <stdio.h>
#include <stdlib.h>
#include "Instruction_Reduced_Reader.hh"
#include "types-and-constants.hh"
// ===================================================================



// ===================================================================
Instruction_Reduced_Reader::Instruction_Reduced_Reader
(const char* const reference_trace_pathname) :
  Reference_Reader(reference_trace_pathname) {

  // Reset the clocks the zero.
  instruction_time = 0;
  reference_time = 0;

  // Reset the previous record values to zero, as the difference
  // encoder will have used that same starting point.
  previous_cycle_timestamp = 0;
  previous_context_ID = 0;

  // An array storing the previous virtual addresses of each type,
  // initialized.  Note that for efficiency and simplicity, we
  // allocate an array that is large enough so that there is an entry
  // for each possible tag value.
  const unsigned int bits_per_byte = 8;
  const unsigned int number_possible_tags =
    (unsigned int)pow(2.0, (double)(sizeof(tag_t) * bits_per_byte));
  previous_address_table = new virtual_address_t[number_possible_tags];
  if (previous_address_table == 0) {
    fprintf(stderr,
	    "Instruction_Reduced_Reader(): Failed address table allocation\n");
    exit(1);
  }
  for (unsigned int i = 0; i < number_possible_tags; i++) {
    previous_address_table[i] = 0;
  }

} // Instruction_Reduced_Reader
// ===================================================================



// ===================================================================
Instruction_Reduced_Reader::~Instruction_Reduced_Reader () {

  // Delete the previous address table.
  delete previous_address_table;
  previous_address_table = 0;

} // ~Instruction_Reduced_Reader
// ===================================================================



// ===================================================================
void
Instruction_Reduced_Reader::read
(reference_record_s* const reference_record) {

  // Attempt to read the next record of the reference trace.
  char* read_result = fgets(buffer, BUFFER_SIZE, reference_stream);

  // Did the read fail?
  if (read_result == NULL) {

    // Distinguish between an EOF and other errors, but abort
    // processing in either case.
    if (feof(reference_stream)) {
      fprintf(stderr,
	      "Instruction_Reduced_Reader::read(): Premature EOF\n");
    } else {
      fprintf(stderr,
	      "Instruction_Reduced_Reader::read(): Read error\n");
    }
    exit(1);
  }

  // Parse the record type, which should be the first character of the
  // buffer.
  reference_record->tag = buffer[0];

  // Attempt to parse and process the rest of the record based on its
  // type.  Declare and assign some variables here to avoid compiler
  // complaints of skipping initialization in switch-case statement.
  unsigned int parse_result = 0;
  unsigned int items_to_parse = 0;
  timestamp_t timestamp_difference;
  context_ID_t context_ID;
  char address_difference_sign;
  virtual_address_t address_difference_magnitude;
  timestamp_t current_instruction_time = instruction_time;
  timestamp_t current_reference_time = reference_time;
  switch (reference_record->tag) {

  case TAG_USER_INSTRUCTION_FETCH:
  case TAG_KERNEL_INSTRUCTION_FETCH:

    // Tick the instruction clock to reflect that _after_ this event,
    // instruction time will have advanced.
    instruction_time++;

    // Fall through to read and parse the rest of the record...

  case TAG_USER_READ:
  case TAG_KERNEL_READ:
  case TAG_USER_WRITE:
  case TAG_KERNEL_WRITE:

    // Tick the reference clock to reflect that _after_ this event,
    // reference time will have advanced.
    reference_time++;

    // Read the remainder of the record.
    items_to_parse = 5;
    parse_result = sscanf(&buffer[2],
			  "%qx %hx %lx %c %lx",
			  &timestamp_difference,
			  &reference_record->length,
			  &context_ID,
			  &address_difference_sign,
			  &address_difference_magnitude);

    // Place the current instruction and reference times into the
    // structure.
    reference_record->instruction_timestamp = current_instruction_time;
    reference_record->reference_timestamp = current_reference_time;

    // Calculate the cycle time based on the previous timestamp and
    // the timestamp difference.
    previous_cycle_timestamp += timestamp_difference;
    reference_record->cycle_timestamp = previous_cycle_timestamp;

    // Calculate the context ID.  If the value read is zero, then it
    // is the same as the previous context ID; otherwise, the value
    // read is the new context ID.
    if (context_ID == 0) {
      reference_record->virtual_page.context_ID = previous_context_ID;
    } else {
      previous_context_ID = context_ID;
      reference_record->virtual_page.context_ID = context_ID;
    }

    // Calculate the virtual address based on the previous address and
    // the address difference.  Virtual addresses are calculated on a
    // per-reference-type basis.
    if (address_difference_sign == '+') {

      previous_address_table[reference_record->tag] +=
	address_difference_magnitude;

    } else if (address_difference_sign == '-') {

      previous_address_table[reference_record->tag] -=
	address_difference_magnitude;

    } else {

      fprintf(stderr, "Instruction_Reduced_Reader::read(): ");
      fprintf(stderr,
	      "Unknown tag %c(%d) at reference time = %qx\n",
	      address_difference_sign,
	      (int)address_difference_sign,
	      current_reference_time);
      exit(1);

    }
    reference_record->virtual_address =
      previous_address_table[reference_record->tag];

    // Compute the page ID by isolating the upper bits of the virtual
    // address and recording that value in the structure.
    reference_record->virtual_page.page_ID =
      reference_record->virtual_address >> PAGE_SIZE_ORDER;

    break;

  case TAG_KERNEL_EVENT:

    // Read the remainder of the record.  We care only about the
    // timestamp.
    items_to_parse = 1;
    parse_result = sscanf(&buffer[2],
			  "%qx",
			  &timestamp_difference);

    // Calculate the cycle time based on the previous timestamp and
    // the timestamp difference.
    previous_cycle_timestamp += timestamp_difference;
    reference_record->cycle_timestamp = previous_cycle_timestamp;

    // Assign the rest of the fields.
    reference_record->instruction_timestamp = current_instruction_time;
    reference_record->reference_timestamp = current_reference_time;
    reference_record->length = 0;
    reference_record->virtual_page.context_ID = 0;
    reference_record->virtual_page.page_ID = 0;
    reference_record->virtual_address = 0;
    break;

  case TAG_END_OF_TRACE:

    // Nothing to do.
    break;

  default:
    fprintf(stderr, "Instruction_Reduced_Reader::read(): ");
    fprintf(stderr,
	    "Invalid reference tag %c(%d)\n",
	    reference_record->tag,
	    (int)reference_record->tag);
    exit(1);

  }

  // If the parse failed, emit the error and exit.
  if (parse_result != items_to_parse) {
    fprintf(stderr, "Instruction_Reduced_Reader::read(): ");
    fprintf(stderr, "Could not parse remainder\n");
    fprintf(stderr,
	    "cycle_timestamp = %qx\n",
	    reference_record->cycle_timestamp);
    fprintf(stderr,
	    "instruction_timestamp = %qx\n",
	    reference_record->instruction_timestamp);
    fprintf(stderr,
	    "reference_timestamp = %qx\n",
	    reference_record->reference_timestamp);
    exit(1);
  }

} // read
// ===================================================================
