/**
 * \file Instruction_Reduced_Reader.hh
 * \author Scott F. H. Kaplan <sfkaplan@cs.amherst.edu>
 * \date January 2003
 * \brief Header of a class that reads and parses instruction-level
 *        reduced reference traces.
 *
 * A subclass of Reference_Reader that can read the instruction-level
 * reduced reference traces produced by the
 * <i>instruction-level-reducer</i> reference handler.  The data from
 * each record is then transformed appropriately for use by
 * Laplace-merge.
 **/



// ===================================================================
// Avoid multiple inclusion.
#if !defined (_INSTRUCTION_REDUCED_READER_HH)
#define _INSTRUCTION_REDUCED_READER_HH
// ===================================================================



// ===================================================================
// INCLUDES

#include <stdio.h>
#include "Reference_Reader.hh"
// ===================================================================



// ===================================================================
// CLASS DEFINITION

/** \brief A subclass of Reference_Reader that reads instruction-level
 *         reduced reference traces.
 *
 * It defines only the read() method that performs the actual reading,
 * parsing, and processing of each reference record before it is
 * provided to the caller in a reference_record_s structure.
 **/
class Instruction_Reduced_Reader : public Reference_Reader {

public: // CONSTRUCTOR & DESTRUCTOR

  /**
   * \brief The constructor.
   * \param reference_trace_pathname The pathname of the reference
   *        trace to be read.
   *
   * Set up an Instruction_Reduced_Reader by initializing data members
   * and by opening the reference trace file and holding a stream for
   * it.
   **/
  Instruction_Reduced_Reader (const char* const reference_trace_pathname);

  /**
   * \brief The destructor.
   *
   * Deallocate the table used to track tag-indexed previous virtual
   * addresses for difference decoding.
   **/
  virtual ~Instruction_Reduced_Reader ();

public: // VIRTUAL METHODS

  /**
   * \brief Read the next record from the reference trace.
   * \see reference_record_s
   * \param reference_record A data structure that will hold the
   *        needed data from the trace record.
   *
   * Read a single record, parse its data, process the values, and
   * place the needed results in the provided structure.  Any errors
   * in reading (I/O errors or premature EOF) are expected to abort
   * processing.
   **/
  virtual void read (reference_record_s* const reference_record);

protected: // DATA

  /** \brief The current reference time. */
  timestamp_t reference_time;

  /** \brief The current instruction time. */
  timestamp_t instruction_time;

  /** \brief The cycle timestamp of the previous record, used to
   *         calculate the absolute cycle time. */
  timestamp_t previous_cycle_timestamp;

  /** \brief The context ID of the previous record, used to calculate
   *         the current context ID. */
  context_ID_t previous_context_ID;

  /** \brief An array that stores the virtual address of the previous
   *         record, indexed by tag (that is, reference type). */
  virtual_address_t* previous_address_table;

}; // class Reference_Reader
// ===================================================================



// ===================================================================
#endif // _INSTRUCTION_REDUCED_READER_HH
// ===================================================================
