/**
 * \file Page_Reduced_Reader.cc
 * \author Scott F. H. Kaplan <sfkaplan@cs.amherst.edu>
 * \date January 2003
 * \brief Code for a class that reads and parses instruction-level
 * reduced reference traces.
 *
 * A subclass of Reference_Reader that can read the page-level reduced
 * reference traces produced by the <i>page-level-reducer</i>
 * reference handler.  The data from each record is then transformed
 * appropriately for use by Laplace-merge.
 */



// ===================================================================
// INCLUDES

#include <stdio.h>
#include <stdlib.h>
#include "Page_Reduced_Reader.hh"
#include "types-and-constants.hh"
// ===================================================================



// ===================================================================
// CONSTANTS

/** \brief The filename of the task ID list. */
#define TASK_ID_LIST_FILENAME "task-ID-list"

/** \brief The filename of the reduction memory size file. */
#define REDUCTION_MEMORY_SIZE_FILENAME "SAD-reduction-memory-size"
// ===================================================================



// ===================================================================
// Don't use the superclass constructor, as it's not relevant here.
Page_Reduced_Reader::Page_Reduced_Reader
(const char* const reference_trace_directory_pathname) {

  // Ensure that the inhereted single reference stream is nullified.
  reference_stream = NULL;

  // Construct the pathnames for the reduction memory size file and
  // the task ID list file.
  char reduction_memory_size_pathname[BUFFER_SIZE];
  sprintf(reduction_memory_size_pathname,
	  "%s/%s",
	  reference_trace_directory_pathname,
	  REDUCTION_MEMORY_SIZE_FILENAME);
  char task_ID_list_pathname[BUFFER_SIZE];
  sprintf(task_ID_list_pathname,
	  "%s/%s",
	  reference_trace_directory_pathname,
	  TASK_ID_LIST_FILENAME);

  // Attempt to open these two files, checking that they opened
  // successfully.
  FILE* const reduction_memory_size_stream =
    fopen(reduction_memory_size_pathname, "r");
  if (reduction_memory_size_stream == NULL) {
    perror("Failed to open reduction memory size file");
    exit(1);
  }
  FILE* const task_ID_list_stream = fopen(task_ID_list_pathname, "r");
  if (task_ID_list_stream == NULL) {
    perror("Failed to open task ID list file");
    exit(1);
  }

  // Read the reduction memory size, ensuring success.
  unsigned int reduction_memory_size;
  int read_result = fscanf(reduction_memory_size_stream,
			   "%d",
			   &reduction_memory_size);
  if (read_result != 1) {
    fprintf(stderr, "Page_Reduced_Reader::Page_Reduced_Reader(): ");
    fprintf(stderr, "Failed to read reduction memory size\n");
    exit(1);
  }

  // Close the reduction memory size file.
  int close_result = fclose(reduction_memory_size_stream);
  if (close_result == EOF) {
    perror("Failed to close reduction memory size file");
    exit(1);
  }

  // Loop through entries in the task ID list, each time opening the
  // corresponding reference trace file and placing an entry in the
  // table that maps task IDs to reference trace streams.
  bool done = false;
  do {

    // Attempt to read the next entry in the task ID list.
    task_ID_t task_ID;
    int read_result = fscanf(task_ID_list_stream,
			     "%hd\n",
			     &task_ID);

    // Check the result of the read?
    if (read_result == 0) {

      // Failure to parse.
      fprintf(stderr, "Page_Reduced_Reader::Page_Reduced_Reader(): ");
      fprintf(stderr, "Failed parse of task ID\n");
      exit(1);

    } else if (read_result == EOF) {

      // An I/O error.  If it's an EOF, that's normal; other kinds of
      // errors are not.
      if (feof(task_ID_list_stream)) {
	done = true;
      } else {
	fprintf(stderr, "Page_Reduced_Reader::Page_Reduced_Reader(): ");
	fprintf(stderr, "Failed read of task ID list file\n");
	exit(1);
      }

    } else {

      // The read succeeded.  Create a per-task reader for this task
      // ID.
      Per_Task_Page_Reduced_Reader* per_task_reader =
	new Per_Task_Page_Reduced_Reader(reference_trace_directory_pathname,
					 reduction_memory_size,
					 task_ID);

      // Add a list entry to associate this task ID with its reader.
      task_reader_pair_t* task_reader_pair =
	new task_reader_pair_t(task_ID, per_task_reader);
      task_reader_pair_queue.push(task_reader_pair);

    }

  } while (!done);

  // Close the task ID list file.
  close_result = fclose(task_ID_list_stream);
  if (close_result == EOF) {
    perror("Failed to close task ID list file");
    exit(1);
  }

} // Page_Reduced_Reader
// ===================================================================



// ===================================================================
Page_Reduced_Reader::~Page_Reduced_Reader () {

  // Sanity check:  It should be the case that all of the reference
  // traces have been completed consumed, and thus all per-task
  // readers should have been eliminated.
  if (!task_reader_pair_queue.empty()) {
    fprintf(stderr, "Page_Reduced_Reader::~Page_Reduced_Reader(): ");
    fprintf(stderr, "Per-task readers remaining\n");
    exit(1);
  }

} // ~Page_Reduced_Reader
// ===================================================================



// ===================================================================
void
Page_Reduced_Reader::read
(reference_record_s* const reference_record) {

  // Are there any open traces with records still to read?
  if (task_reader_pair_queue.empty()) {

    // There are no more records to read from any trace.  Return an
    // end-of-trace record and a value that indicates that no
    // reference records were available.
    reference_record->tag = TAG_END_OF_TRACE;

  } else {

    // There is at least one more record in some task's trace to read.
    // Grab the task/reader pair at the top of the queue (and thus the
    // one with the smallest timestamp).  Remove it from the queue.
    task_reader_pair_t* smallest_task_reader_pair =
      task_reader_pair_queue.top();
    Per_Task_Page_Reduced_Reader* per_task_reader =
      smallest_task_reader_pair->second;
    task_reader_pair_queue.pop();

    // Grab the record from this task's reader.
    per_task_reader->get_record(reference_record);

    // Have we reached the end of this trace?  (That is, does the next
    // record indicate an end-of-trace for this task?)
    if (per_task_reader->at_end_of_trace()) {

      // There are no more records in this trace.  Delete the per task
      // reader and the task/reader pair.
      delete per_task_reader;
      per_task_reader = 0;
      delete smallest_task_reader_pair;
      smallest_task_reader_pair = 0;

    } else {

      // There remain more records in this trace.  Re-insert the
      // task/reader pair into the priority queue.
      task_reader_pair_queue.push(smallest_task_reader_pair);

    }
  }

} // read
// ===================================================================
