/**
 * \file Page_Reduced_Reader.hh
 * \author Scott F. H. Kaplan <sfkaplan@cs.amherst.edu>
 * \date January 2003
 * \brief Header of a class that reads and parses page-level
 *        reduced reference traces.
 *
 * A subclass of Reference_Reader that can read the page-level reduced
 * reference traces produced by the <i>page-level-reducer</i>
 * reference handler.  The data from each record is then transformed
 * appropriately for use by Laplace-merge.
 **/



// ===================================================================
// Avoid multiple inclusion.
#if !defined (_PAGE_REDUCED_READER_HH)
#define _PAGE_REDUCED_READER_HH
// ===================================================================



// ===================================================================
// INCLUDES

#include <stdio.h>
#include <stack.h> // For priority_queue, oddly enough.
#include "Per_Task_Page_Reduced_Reader.hh"
#include "Reference_Reader.hh"
// ===================================================================



// ===================================================================
// TYPES

/** \brief A type that encapsulates a task ID and its per-task
 *         reader. */
typedef pair<task_ID_t,
	     Per_Task_Page_Reduced_Reader*> task_reader_pair_t;

/**
 * \brief A type that defines a comparison function between
 *        task/reader pairs.
 *
 * The definition of an STL priority queue is that it uses a function
 * like this one to return true if x < y, and then return the largest
 * element in the queue.  By defining this function to return true if
 * x > y, we force the priority queue to return the smallest element
 * in the queue.
 **/
struct task_reader_greater_s {
  bool operator() (const task_reader_pair_t* const x,
		   const task_reader_pair_t* const y) {
    return (x->second->next_cycle_timestamp() >
	    y->second->next_cycle_timestamp());
  }
};

/** \brief A priority queue type that stores and orders task
 *         ID/per-task-readers pairs. */
typedef priority_queue<task_reader_pair_t*,
		       vector<task_reader_pair_t*>,
		       task_reader_greater_s> task_reader_pair_queue_t;
// ===================================================================



// ===================================================================
// CLASS DEFINITION

/**
 * \brief A subclass of Reference_Reader that reads page-level
 *        reduced reference traces.
 *
 * It defines only the read() method that performs the actual reading,
 * parsing, and processing of each reference record before it is
 * provided to the caller in a reference_record_s structure.
 **/
class Page_Reduced_Reader : public Reference_Reader {

public: // CONSTRUCTOR & DESTRUCTOR

  /**
   * \brief The constructor.
   * \param reference_trace_directory_pathname The pathname of the
   *        <b>directory</b> of reference traces to be read.
   *
   * Set up a Page_Reduced_Reader by initializing data members.  Also
   * open all of the reference trace files (with some help from other
   * files in the same directory) and build a priority queue that
   * assocaites task IDs with reference trace streams, ordering them
   * by earliest reference record.
   **/
  Page_Reduced_Reader
  (const char* const reference_trace_directory_pathname);

  /**
   * \brief The destructor.
   *
   * Close all input streams and deallocate any dynamically allocated
   * data members.
   **/
  virtual ~Page_Reduced_Reader ();

public: // VIRTUAL METHODS

  /**
   * \brief Read the next record from the reference trace.
   * \see reference_record_s
   * \param reference_record A data structure that will hold the
   *        needed data from the trace record.
   *
   * Read a single record, parse its data, process the values, and
   * place the needed results in the provided structure.  From the
   * multiple open reference traces, select the reference that occurs
   * first according to the cycle timestamps.  Any errors in reading
   * (I/O errors or premature EOF) are expected to abort processing.
   **/
  virtual void read (reference_record_s* const reference_record);

protected: // DATA

  /**
   * \brief A queue that stores and orders task ID/per-task-reader
   *        pairs.
   *
   * A priority queue that holds pairs which associate a task ID with
   * its per-task reference trace reader.  The pairs are ordered by
   * the timestamp of the next record, where obtaining the element in
   * the queue is the one with the smallest timestamp.
   **/
  task_reader_pair_queue_t task_reader_pair_queue;

}; // class Reference_Reader
// ===================================================================



// ===================================================================
#endif // _PAGE_REDUCED_READER_HH
// ===================================================================
