/**
 * \file Per_Task_Consumer.cc
 * \author Scott F. H. Kaplan -- sfkaplan@cs.amherst.edu
 * \date November 2002
 * \brief Produces per-task reference files, grouped by process.
 **/



// ===================================================================
// INCLUDES

#include <stdio.h>
#include <string.h>
#include "Per_Task_Consumer.hh"
#include "PTC_Task.hh"
// ===================================================================



// ===================================================================
Per_Task_Consumer::Per_Task_Consumer
(Reference_Reader* const reference_reader,
 const char* const kernel_pathname,
 const char* const output_directory_pathname) :
  Consumer(reference_reader, kernel_pathname) {

  // Copy the output directory pathname into this object.
  strcpy(this->output_directory_pathname, output_directory_pathname);

  // Create a writer for the kernel task.
  PTC_Task* task = (PTC_Task*)kernel_task;
  task->writer = new Per_Task_Writer(output_directory_pathname,
				     task->process->process_ID,
				     task->task_ID);
  if (task->writer == NULL) {
    fprintf(stderr, "Per_Task_Consumer::Per_Task_Consumer(): ");
    fprintf(stderr, "Failed to allocate kernel task's writer\n");
    exit(1);
  }

} // Per_Task_Consumer::Per_Task_Consumer
// ===================================================================



// ===================================================================
Per_Task_Consumer::~Per_Task_Consumer () {

  // Traverse the task_ID->Task table.
  task_ID_to_Task_map_t::iterator i = task_ID_to_Task_map.begin();
  while (i != task_ID_to_Task_map.end()) {

    // Delete the per-task writer associated with this task, thus
    // flushing and closing its output.
    PTC_Task* task = (PTC_Task*)(*i).second;
    delete task->writer;
    task->writer = NULL;

    // Move to the next task.
    i++;

  } // task_ID->Task table traversal

} // Per_Task_Consumer::~Per_Task_Consumer
// ===================================================================



// ===================================================================
void
Per_Task_Consumer::act_on_reference_record () {

  // Lookup the canonical page ID, if one exists, for this reference
  // in the scheduled task's context.  Note that we always use the map
  // of the scheduled task, as the kernel's mapping is static (and
  // doesn't require a table), and any references below the kernel's
  // address range will be through the scheduled process's map.
  bool V2C_mapping_exists;
  canonical_page_ID_t canonical_page_ID;
  lookup_canonical_page_ID(&V2C_mapping_exists,
			   &canonical_page_ID,
			   scheduled_task->process->virtual_to_canonical_map,
			   reference_record->virtual_page.page_ID);

  // Submit this reference to this task's writer only if a mapping to
  // a canonical page really exists (that is, this is not a reference
  // to a buffer-cache page).
  if (V2C_mapping_exists) {
    PTC_Task* active_PTC_task = (PTC_Task*)active_task;
    active_PTC_task->writer->
      append_reference_record(reference_record,
			      canonical_page_ID,
			      active_task->virtual_cycle_time,
			      active_task->virtual_instruction_time,
			      active_task->virtual_reference_time);
  }

} // Consumer::act_on_reference_record
// ===================================================================



// ===================================================================
void
Per_Task_Consumer::act_on_kernel_record () {

  // Determine how to react to this record depending on the type of
  // kernel event.
  PTC_Task* active_PTC_task = NULL;
  switch (kernel_record->tag) {

  case TAG_CONTEXT_ASSIGNMENT:

    PTC_handle_context_assignment();
    break;

  case TAG_FILE_OPEN:
  case TAG_FILE_CLOSE:
  case TAG_FILE_READ:
  case TAG_FILE_WRITE:
  case TAG_ACCEPT:

    // These kernel events should appear in the output trace for the
    // current task.  As a sanity check, ensure that the task ID given
    // for this record matches the current task's ID.
    if (kernel_record->task_ID != scheduled_task->task_ID) {
      fprintf(stderr, "Per_Task_Consumer::act_on_kernel_record(): ");
      fprintf(stderr, "Record and scheduled task ID mismatch\n");
      exit(1);
    }

    // Pass the kernel event to the task's writer.
    active_PTC_task = (PTC_Task*)active_task;
    active_PTC_task->writer->
      append_kernel_record(kernel_record,
			   active_task->virtual_cycle_time,
			   active_task->virtual_instruction_time,
			   active_task->virtual_reference_time);
    break;

  default:

    // Do nothing -- ignore this kernel record, as it's not relevant
    // to the output of these per-task traces.
    break;

  }

} // Consumer::act_on_kernel_record
// ===================================================================



// ===================================================================
Task*
Per_Task_Consumer::allocate_new_task () {

  // Attempt to allocate a PTC_Task, a specific subclass designed for
  // use with this class.
  Task* task = new PTC_Task;
  if (task == NULL) {
    fprintf(stderr, "Per_Task_Consumer::allocate_new_task(): ");
    fprintf(stderr, "Allocation failure\n");
    exit(1);
  }

  return task;

} // Per_Task_Consumer::allocate_new_task
// ===================================================================



// ===================================================================
void
Per_Task_Consumer::PTC_handle_context_assignment () {

  // Look up the task being assigned a context.
  PTC_Task* task =
    (PTC_Task*)safe_task_ID_to_task(kernel_record->task_ID);

  // Was there a previous writer?  (The existence of one implies that
  // the task was previously assigned to some other context, and thus
  // some other process.)
  if (task->writer != NULL) {

    // Delete (and thus flush and close) the old writer.
    delete task->writer;
    task->writer = NULL;

  } // Task had previous writer

  // Create a new writer for this task and its newly assigned context
  // and process.
  task->writer = new Per_Task_Writer(output_directory_pathname,
				     task->process->process_ID,
				     task->task_ID);
  if (task->writer == NULL) {
    fprintf(stderr,
	    "Per_Task_Consumer::PTC_handle_context_assignment(): ");
    fprintf(stderr, "Failed to allocate new per-task writer\n");
    exit(1);
  }

} // Per_Task_Consumer::PTC_handle_context_assignment
// ===================================================================
