/**
 * \file Per_Task_Consumer.hh
 * \author Scott F. H. Kaplan -- sfkaplan@cs.amherst.edu
 * \date November 2002
 * \brief Produces per-task reference files, grouped by process.
 *
 * A subclass of Consumer that divides the stream of reference and
 * kernel records on a per-task basis, and groups tasks into
 * processes.  One output file is created for each task, where that
 * file will contain every reference performed by that task as well as
 * every <i>interesting</i> system call (e.g. blocking I/O or file
 * operations).  For each process (that is, group of 1 or more tasks
 * sharing a virtual address space), a separate directory is created
 * and the output files for its tasks are placed in that directory.
 **/



// ===================================================================
// Avoid multiple inclusion.
#if !defined (_PER_TASK_CONSUMER_HH)
#define _PER_TASK_CONSUMER_HH
// ===================================================================



// ===================================================================
// INCLUDES

#include "Consumer.hh"
// ===================================================================



// ===================================================================
// CLASS DEFINITION

/**
 * \brief Divide reference streams by process and task.
 *
 * Produce multiple traces grouped by process and, within each
 * process, by thread (a.k.a. task).
 **/
class Per_Task_Consumer : public Consumer {

public: // CONSTRUCTOR

  /**
   * \brief The constructor.
   * \param reference_reader The object that will read and parse the
   *        reference trace.
   * \param kernel_pathname The pathname of the kernel trace to be
   *        read.
   * \param output_directory_pathname The pathname of the directory
   *        that will hold all of the output files.
   *
   * Call on the superclass to open the kernel trace.  Store the
   * output directory pathname for later, when tasks and processes are
   * encountered and require new directories and files to be created.
   **/
  Per_Task_Consumer (Reference_Reader* const reference_reader,
		     const char* const kernel_pathname,
		     const char* const output_directory_pathname);

  /**
   * \brief The deconstructor.
   *
   * Close all per-task writers.
   **/
  virtual ~Per_Task_Consumer ();

protected: // VIRTUAL METHODS

  /** \brief Called before processing begins -- Does nothing. */
  virtual void initialize () {}

  /** \brief Called after processing ends -- Does nothing. */
  virtual void clean_up () {}

  /**
   * \brief Called after processing a reference record.
   *
   * Direct the reference to its task's writer, emitting the reference
   * with the logical page number included.
   **/
  virtual void act_on_reference_record ();

  /** \brief Called after processing a kernel record.
   *
   * If it is a blocking I/O or filesystem call, emit it to the task's
   * writer.  If the record indicates a fork, exit, or context
   * assignment, update the per-task writers and mappings.
   **/
  virtual void act_on_kernel_record ();

  /**
   * \brief Allocate a new PTC task object.
   * \return A pointer to a newly-allocated task object.
   *
   * Allocate a new task object, where the object is a PTC_Task
   * object, designed to additionally associate a task with its
   * writer.  Abort if the allocation fails.
   **/
  virtual Task* allocate_new_task ();

protected: // SUPPORT METHODS

  /** \brief Close an old file (if appropriate) and open a new one to
   *         reflect the beginning of a  */
  void PTC_handle_context_assignment ();

protected: // DATA MEMBERS

  /** \brief The directory into which all results will be placed. */
  char output_directory_pathname[BUFFER_SIZE];

}; // class Per_Task_Consumer
// ===================================================================



// ===================================================================
#endif // _PER_TASK_CONSUMER_HH
// ===================================================================
