/**
 * \file Per_Task_Page_Reduced_Reader.cc
 * \author Scott F. H. Kaplan <sfkaplan@cs.amherst.edu>
 * \date January 2003
 * \brief Header of a class that reads and parses a single task's
 *        page-level reduced reference trace.
 *
 * The methods of the class that can read a single page-level reduced
 * reference trace created by the <i>page-level reduction</i>
 * reference handler.
 **/



// ===================================================================
// INCLUDES

#include <stdio.h>
#include "Per_Task_Page_Reduced_Reader.hh"
// ===================================================================



// ===================================================================
Per_Task_Page_Reduced_Reader::Per_Task_Page_Reduced_Reader
(const char* const reference_trace_directory_pathname,
 const unsigned int reduction_memory_size,
 const task_ID_t task_ID) {

  // Construct the pathname of the reference trace corresponding to
  // this task ID.
  char reference_trace_pathname[BUFFER_SIZE];
  sprintf(reference_trace_pathname,
	  "%s/%hd.sad-%d.gz",
	  reference_trace_directory_pathname,
	  task_ID,
	  reduction_memory_size);

  // Open this task's reference trace.
  input_stream = gzopen(reference_trace_pathname, "r");
  if (input_stream == NULL) {
    perror("Failed to open reference stream");
    exit(1);
  }

  // Read the first record, thus making the timestamp of the next
  // record available.
  read_record();

} // Page_Reduced_Reader
// ===================================================================



// ===================================================================
Per_Task_Page_Reduced_Reader::~Per_Task_Page_Reduced_Reader () {

  // Close the input stream, ensuring success.
  int close_result = gzclose(input_stream);
  if (close_result != Z_OK) {
    fprintf(stderr, "Per_Task_Page_Reduced_Reader::~Per_Task_Page_Reduced_Reader(): ");
    fprintf(stderr, "Failed close of trace file\n");
    exit(1);
  }

} // ~Per_Task_Page_Reduced_Reader
// ===================================================================



// ===================================================================
void
Per_Task_Page_Reduced_Reader::get_record
(reference_record_s* const reference_record) {

  // Copy the current record into the provided structure.
  *reference_record = this->reference_record;

  // Read the next record from the trace file.
  read_record();

} // Per_Task_Page_Reduced_Reader::get_record
// ===================================================================



// ===================================================================
bool
Per_Task_Page_Reduced_Reader::at_end_of_trace () const {

  return (reference_record.tag == TAG_END_OF_TRACE);

} // Per_Task_Page_Reduced_Reader::at_end_of_trace
// ===================================================================



// ===================================================================
timestamp_t
Per_Task_Page_Reduced_Reader::next_cycle_timestamp () const {

  return reference_record.cycle_timestamp;

} // Per_Task_Page_Reduced_Reader::next_timestamp
// ===================================================================



// ===================================================================
void
Per_Task_Page_Reduced_Reader::read_record () {

  // Read the next record, ensuring success.
  char buffer[BUFFER_SIZE];
  char* read_result = gzgets(input_stream, buffer, BUFFER_SIZE);
  if (read_result == Z_NULL) {
    fprintf(stderr, "Per_Task_Page_Reduced_Reader::read_record(): ");
    fprintf(stderr, "Failed read\n");
    exit(1);
  }

  // Determine how to parse this record based on its type.  If it an
  // end-of-trace record, then there's nothing left to do.  Otherwise,
  // read the full entry.
  reference_record.tag = buffer[0];
  if (reference_record.tag != TAG_END_OF_TRACE) {

    // Attempt to parse the remainder of the record.
    timestamp_t cycle_timestamp_difference;
    timestamp_t instruction_timestamp_difference;
    timestamp_t reference_timestamp_difference;
    context_ID_t context_ID_input;
    char page_ID_difference_sign;
    virtual_page_ID_t page_ID_difference_magnitude;
    int items_to_parse = 6;
    int read_result = sscanf(&buffer[2],
			     "%qx %qx %qx %lx %c %lx",
			     &cycle_timestamp_difference,
			     &instruction_timestamp_difference,
			     &reference_timestamp_difference,
			     &context_ID_input,
			     &page_ID_difference_sign,
			     &page_ID_difference_magnitude);

    // Ensure that the parse succeeded.
    if (read_result != items_to_parse) {
      fprintf(stderr, "Per_Task_Page_Reduced_Reader::read_record(): ");
      fprintf(stderr, "Failed parse\n");
      exit(1);
    }

    // Calculate the absolute timestamp values.
    reference_record.cycle_timestamp =
      previous_cycle_timestamp + cycle_timestamp_difference;
    reference_record.instruction_timestamp =
      previous_instruction_timestamp + instruction_timestamp_difference;
    reference_record.reference_timestamp =
      previous_reference_timestamp + reference_timestamp_difference;

    // Update the previous timestamps for later.
    previous_cycle_timestamp = reference_record.cycle_timestamp;
    previous_instruction_timestamp =
      reference_record.instruction_timestamp;
    previous_reference_timestamp = reference_record.reference_timestamp;

    // If the record indicated a context ID of 0, then it implies that
    // the context ID is the same as the last record; otherwise, it is
    // a new context ID, different from the last record's.
    if (context_ID_input == 0) {
      reference_record.virtual_page.context_ID = previous_context_ID;
    } else {
      reference_record.virtual_page.context_ID = context_ID_input;
      previous_context_ID = context_ID_input;
    }

    // Calculate the absolute page ID based on the difference from the
    // previous record.
    switch (page_ID_difference_sign) {

    case '+':
      reference_record.virtual_page.page_ID =
	previous_page_ID + page_ID_difference_magnitude;
      break;

    case '-':
      reference_record.virtual_page.page_ID =
	previous_page_ID - page_ID_difference_magnitude;
      break;

    default:
      fprintf(stderr, "Per_Task_Page_Reduced_Reader::read_record(): ");
      fprintf(stderr,
	      "Invalid difference sign %c(%d)\n",
	      page_ID_difference_sign,
	      (unsigned int)page_ID_difference_sign);
      exit(1);

    }
    previous_page_ID = reference_record.virtual_page.page_ID;

    // Set the length of the reference to 0, as this value is
    // meaningless in this context.
    reference_record.length = 0;

    // Set the virtual address referenced to the first byte of the
    // indicated page.
    reference_record.virtual_address =
      reference_record.virtual_page.page_ID << PAGE_SIZE_ORDER;

  }

} // Per_Task_Page_Reduced_Reader::read_record
// ===================================================================
