/**
 * \file Per_Task_Page_Reduced_Reader.hh
 * \author Scott F. H. Kaplan <sfkaplan@cs.amherst.edu>
 * \date January 2003
 * \brief Header of a class that reads and parses a single task's
 *        page-level reduced reference trace.
 *
 * A class that can read a single page-level reduced reference trace
 * produced by the <i>page-level-reducer</i> reference handler.  It
 * can read and parse each record, and provide an interface that
 * allows a user to negotiate the interleaving of references from many
 * such traces via multiple PTPRR objects.
 **/



// ===================================================================
// Avoid multiple inclusion.
#if !defined (_PER_TASK_PAGE_REDUCED_READER_HH)
#define _PER_TASK_PAGE_REDUCED_READER_HH
// ===================================================================



// ===================================================================
// INCLUDES

#include <zlib.h>
#include "types-and-constants.hh"
// ===================================================================



// ===================================================================
// CLASS DEFINITION

/**
 * \brief A class that reads a single page-level reduced reference
 *        trace.
 *
 * Read and parse the records of a single page-level reduced reference
 * trace.  With each record provided to the caller, read ahead to the
 * next record so that the time of the next reference can also be
 * available to the caller.
 **/
class Per_Task_Page_Reduced_Reader {

public: // CONSTRUCTOR & DESTRUCTOR

  /**
   * \brief The constructor.
   * \param reference_trace_directory_pathname The pathname of the
   *        <b>directory</b> of reference traces to be read.
   * \param reduction_memory_size The memory size on which the trace
   *        was reduced.
   * \param task_ID The ID of the task for which this object will read
   *        references.
   *
   * Set up a Per_Task_Page_Reduced_Reader by opening its reference
   * trace file and reading its first record (so that the timestamp of
   * the next reference can be known).
   **/
  Per_Task_Page_Reduced_Reader
  (const char* const reference_trace_directory_pathname,
   const unsigned int reduction_memory_size,
   const task_ID_t task_ID);

  /**
   * \brief The destructor.
   *
   * Close the input stream.
   **/
  ~Per_Task_Page_Reduced_Reader ();

public: // METHODS

  /**
   * \brief Provide the next reference record from the trace.
   * \param reference_record A data structure that will hold the
   *        needed data from the trace record.
   *
   * Copy the current reference record into the provided structure.
   * Read the <b>next</b> record from the trace, thus making the
   * timestamp of the next record available.
   **/
  void get_record (reference_record_s* const reference_record);

  /** \brief Return the cycle timestamp of the next record. */
  timestamp_t next_cycle_timestamp () const;

  /** \brief Return whether the next record is an end-of-trace
   *         record. */
  bool at_end_of_trace () const;

protected: // METHODS

  /** \brief Read the next record from the trace into an internal
   *         record structure. */
  void read_record ();

protected: // DATA

  /** \brief The compressed input stream through which the trace file
   *         is read.  */
  gzFile input_stream;

  /** \brief The current record from this trace, stored so that its
   *         timestamp can be provided. */
  reference_record_s reference_record;

  /** \brief The cycle timestamp of the previous record, used to
   *         calculate the absolute cycle time. */
  timestamp_t previous_cycle_timestamp;

  /** \brief The instruction timestamp of the previous record, used to
   *         calculate the absolute instruction time. */
  timestamp_t previous_instruction_timestamp;

  /** \brief The reference timestamp of the previous record, used to
   *         calculate the absolute reference time. */
  timestamp_t previous_reference_timestamp;

  /** \brief The context ID of the previous record, used to calculate
   *         the current context ID. */
  context_ID_t previous_context_ID;

  /** \brief The page ID of the previous record, used to calculate the
   *         current page ID. */
  virtual_page_ID_t previous_page_ID;

}; // class Reference_Reader
// ===================================================================



// ===================================================================
#endif // _PER_TASK_PAGE_REDUCED_READER_HH
// ===================================================================
