/**
 * \file Reference_Reader.hh
 * \author Scott F. H. Kaplan <sfkaplan@cs.amherst.edu>
 * \date January 2003
 * \brief Header of base class for reading and parsing reference
 * traces.
 *
 * A class that defines the interface for reading and parsing a
 * reference trace.  This is an ABSTRACT CLASS: It does not define the
 * read() method.  By defining this method, a subclass actually read
 * and parse some specific reference trace format and place the needed
 * results into the data structure that read() accepts as an
 * argument.
 */



// ===================================================================
// Avoid multiple inclusion.
#if !defined (_REFERENCE_READER_HH)
#define _REFERENCE_READER_HH
// ===================================================================



// ===================================================================
// INCLUDES

#include <stdio.h>
#include "types-and-constants.hh"
// ===================================================================




// ===================================================================
// CLASS DEFINITION

/**
 * \brief An abstract class whose subclasses can read different trace
 *        formats.
 *
 * This abstract class is a simple one.  It will open and close the
 * file, and provide the ability to rewind that file.  It does not
 * define how the file is read, but rather provides the method that a
 * subclass must define in order to read a specific reference trace
 * format.
 **/
class Reference_Reader {

public: // CONSTRUCTOR & DESTRUCTOR

  /**
   * \brief The default constructor.
   *
   * Intended for subclasses that do not open a single file, and
   * therefore choose not to use the provided file stream data
   * member.
   **/
  Reference_Reader () {}

  /**
   * \brief The common constructor.
   * \param reference_trace_pathname The pathname of the reference
   *        trace to be read.
   *
   * Set up a Reference_Reader object by opening the reference trace
   * file and holding a stream for it.
   **/
  Reference_Reader (const char* const reference_trace_pathname);

  /**
   * \brief The deconstructor.
   *
   * Deconstruct a Reference_Reader.  Close the reference trace
   * file.
   **/
  virtual ~Reference_Reader ();

public: // PURE VIRTUAL METHODS

  /**
   * \brief Read the next record from the reference trace.
   * \see reference_record_s
   * \param reference_record A data structure that will hold the
   *        needed data from the trace record.
   *
   * This pure virtual method must be defined in a subclass so that it
   * reads the next record, parses it, and places the required
   * information into the data structure passed to it.  Any errors in
   * reading (I/O errors or premature EOF) are expected to abort
   * processing.
   **/
  virtual void read (reference_record_s* const reference_record) = 0;

protected: // DATA

  /** \brief The reference trace input stream. */
  FILE* reference_stream;

  /** \brief A buffer used for input and parsing. */
  char buffer[BUFFER_SIZE];

}; // class Reference_Reader
// ===================================================================



// ===================================================================
#endif // _REFERENCE_READER_HH
// ===================================================================
