/**
 * \file Text_Raw_Reader.cc
 * \author Scott F. H. Kaplan <sfkaplan@cs.amherst.edu>
 * \date January 2003
 * \brief Code for a class that reads and parses text raw reference
 * traces.
 *
 * A subclass of Reference_Reader that can read the text raw reference
 * traces produced by the <i>raw-reference-binary-to-text</i>
 * utility.  The data from each record is then transformed
 * appropriately for use by Laplace-merge.
 */



// ===================================================================
// INCLUDES

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include "Text_Raw_Reader.hh"
#include "types-and-constants.hh"
// ===================================================================



// ===================================================================
Text_Raw_Reader::Text_Raw_Reader
(const char* const reference_trace_pathname) :
  Reference_Reader(reference_trace_pathname) {

  // Reset the clocks to zero.
  reference_time = 0;
  instruction_time = 0;

} // Text_Raw_Reader
// ===================================================================



// ===================================================================
void
Text_Raw_Reader::read (reference_record_s* const reference_record) {

  // Attempt to read the next record of the reference trace.
  char* read_result = fgets(buffer, BUFFER_SIZE, reference_stream);

  // Did the read fail?
  if (read_result == NULL) {

    // Distinguish between an EOF and other errors, but abort
    // processing in either case.
    if (feof(reference_stream)) {
      fprintf(stderr,
	      "Text_Raw_Reader::read(): Premature EOF\n");
    } else {
      fprintf(stderr,
	      "Text_Raw_Reader::read(): Read error\n");
    }
    exit(1);
  }

  // Parse the record type, which should be the first character of the
  // buffer.
  reference_record->tag = buffer[0];

  // Attempt to parse the rest of the record based on its type.
  // Declare and assign some variables here to avoid compiler
  // complaints of skipping initialization in switch-case statement.
  unsigned int parse_result = 0;
  unsigned int items_to_parse = 0;
  timestamp_t current_instruction_time = instruction_time;
  timestamp_t current_reference_time = reference_time;
  switch (reference_record->tag) {

  case TAG_USER_INSTRUCTION_FETCH:
  case TAG_KERNEL_INSTRUCTION_FETCH:

    // Tick the instruction clock to reflect that _after_ this event,
    // instruction time will have advanced.
    instruction_time++;

    // Fall through to read and parse the rest of the record...

  case TAG_USER_READ:
  case TAG_KERNEL_READ:
  case TAG_USER_WRITE:
  case TAG_KERNEL_WRITE:

    // Tick the reference clock to reflect that _after_ this event,
    // reference time will have advanced.
    reference_time++;

    // Read the remainder of the record.
    items_to_parse = 4;
    parse_result = sscanf(&buffer[2],
			  "%qx %hd %lx %lx",
			  &reference_record->cycle_timestamp,
			  &reference_record->length,
			  &reference_record->virtual_page.context_ID,
			  &reference_record->virtual_address);

    // Place the current instruction and reference times into the
    // structure.
    reference_record->instruction_timestamp = current_instruction_time;
    reference_record->reference_timestamp = current_reference_time;

    // Compute the page ID by isolating the upper bits of the virtual
    // address and recording that value in the structure.
    reference_record->virtual_page.page_ID =
      reference_record->virtual_address >> PAGE_SIZE_ORDER;

    break;

  case TAG_KERNEL_EVENT:

    // Read the remainder of the record.  We care only about the
    // timestamp.
    items_to_parse = 1;
    parse_result = sscanf(&buffer[2],
			  "%qx",
			  &reference_record->cycle_timestamp);
    reference_record->instruction_timestamp = current_instruction_time;
    reference_record->reference_timestamp = current_reference_time;
    reference_record->length = 0;
    reference_record->virtual_page.context_ID = 0;
    reference_record->virtual_page.page_ID = 0;
    reference_record->virtual_address = 0;
    break;

  case TAG_END_OF_TRACE:

    // Nothing to do.
    break;

  default:
    fprintf(stderr,
	    "Text_Raw_Reader::read(): Invalid reference tag %c(%d)\n",
	    reference_record->tag,
	    (int)reference_record->tag);
    exit(1);

  }

  // If the parse failed, emit the error and exit.
  if (parse_result != items_to_parse) {
    fprintf(stderr,
	    "Text_Raw_Reader::read(): Could not parse remainder\n");
    fprintf(stderr,
	    "                         cycle_timestamp = %qx\n",
	    reference_record->cycle_timestamp);
    fprintf(stderr,
	    "                         instruction_timestamp = %qx\n",
	    reference_record->instruction_timestamp);
    fprintf(stderr,
	    "                         reference_timestamp = %qx\n",
	    reference_record->reference_timestamp);
    exit(1);
  }

} // read
// ===================================================================
