/**
 * \file Text_Raw_Reader.hh
 * \author Scott F. H. Kaplan <sfkaplan@cs.amherst.edu>
 * \date January 2003
 * \brief Header of a class that reads and parses text raw reference
 * traces.
 *
 * A subclass of Reference_Reader that can read the text raw reference
 * traces produced by the <i>raw-reference-binary-to-text</i>
 * utility.  The data from each record is then transformed
 * appropriately for use by Laplace-merge.
 */



// ===================================================================
// Avoid multiple inclusion.
#if !defined (_TEXT_RAW_READER_HH)
#define _TEXT_RAW_READER_HH
// ===================================================================



// ===================================================================
// INCLUDES

#include <stdio.h>
#include "Reference_Reader.hh"
// ===================================================================



// ===================================================================
// CLASS DEFINITION

/**
 * The Text_Raw_Reader definition.  This subclass of Reference_Reader
 * that reads text raw reference traces.  It defines only the read()
 * method that performs the actual reading, parsing, and processing of
 * each reference record before it is provided to the caller in a
 * reference_record_s structure.
 **/
class Text_Raw_Reader : public Reference_Reader {

public: // CONSTRUCTOR

  /**
   * \brief The constructor.
   * \param reference_trace_pathname The pathname of the reference
   * trace to be read.
   * Set up a Text_Raw_Reader by initializing its reference and
   * instruction clocks and by opening the reference trace file and
   * holding a stream for it.
   **/
  Text_Raw_Reader (const char* const reference_trace_pathname);

public: // VIRTUAL METHODS

  /**
   * \brief Read the next record from the reference trace.
   * \see reference_record_s
   * \param reference_record A data structure that will hold the
   * needed data from the trace record.
   *
   * Read a single record, parse its data, process the values, and
   * place the needed results in the provided structure.  Any errors
   * in reading (I/O errors or premature EOF) are expected to abort
   * processing.
   **/
  virtual void read (reference_record_s* const reference_record);

protected: // DATA

  /** \brief The current reference time. */
  timestamp_t reference_time;

  /** \brief The current instruction time. */
  timestamp_t instruction_time;

}; // class Reference_Reader
// ===================================================================



// ===================================================================
#endif // _TEXT_RAW_READER_HH
// ===================================================================
