// ===================================================================
/**
 * \file main.cc
 * \author Scott F. H. Kaplan <sfkaplan@cs.amherst.edu>
 * \date November 2002
 * \brief The entry point.
 *
 * This file is the entry point for Laplace-merge.  From the command
 * line arguments, it determines what type of merge is requested, what
 * type of reference trace format to parse, creates objects for those
 * choices, and then performs the merge.
 **/
// ===================================================================



// ===================================================================
#include <string.h>
#include "Consumer.hh"
#include "Basic_Consumer.hh"
#include "Per_Task_Consumer.hh"
#include "Reference_Reader.hh"
#include "Binary_Raw_Reader.hh"
#include "Text_Raw_Reader.hh"
#include "Instruction_Reduced_Reader.hh"
#include "Page_Reduced_Reader.hh"
// ===================================================================



// ===================================================================
/**
 * \brief The entry point.
 * 
 * \param argc The number of arguments given at the command line.
 * \param argv The vector of argument strings passed at the command
 *             line.
 * \return An execution result code.
 * 
 * Parse the command line arguments to obtain the type of merge
 * requested and the pathnames of the input traces.  Create and invoke
 * an object that will perform the requested type of merge.
 */

int
main (const unsigned int argc, const char** const argv) {

  // Were the correct number of arguments passed?
  if (argc != 6) {

    // No.  Emit usage and exit.
    fprintf(stderr, "Usage: %s\n", argv[0]);
    fprintf(stderr, "       <merge type (basic, per-task)>\n");
    fprintf(stderr, "       <reference trace type (binary-raw, text-raw, instruction-reduced, page-reduced)>\n");
    fprintf(stderr, "       <reference trace pathname>\n");
    fprintf(stderr, "       <kernel trace pathname>\n");
    fprintf(stderr, "       <output pathname>\n");
    return 1;

  }

  // Assign names to the arguments.
  const char* const merge_type = argv[1];
  const char* const reference_trace_type = argv[2];
  const char* const reference_trace_pathname = argv[3];
  const char* const kernel_trace_pathname = argv[4];
  const char* const output_pathname = argv[5];

  // Determine what kind of reference trace is being used.
  Reference_Reader* reference_reader = 0;
  if (strcmp(reference_trace_type, "text-raw") == 0) {

    reference_reader = new Text_Raw_Reader(reference_trace_pathname);

  } else if (strcmp(reference_trace_type, "binary-raw") == 0) {

    reference_reader =
      new Binary_Raw_Reader(reference_trace_pathname);

  } else if (strcmp(reference_trace_type,
		    "instruction-reduced") == 0) {

    reference_reader =
      new Instruction_Reduced_Reader(reference_trace_pathname);

  } else if (strcmp(reference_trace_type, "page-reduced") == 0) {

    reference_reader =
      new Page_Reduced_Reader(reference_trace_pathname);

  } else {

    fprintf(stderr,
	    "main(): Unknown reference trace type %s\n",
	    reference_trace_type);
    exit(1);

  }

  // Ensure that the reference reader was allocated.
  if (reference_reader == 0) {

    fprintf(stderr, "main(): Unable to allocate reference reader\n");
    exit(1);

  }

  // Determine what kind of consumer was requested.
  Consumer* consumer = 0;
  if (strcmp(merge_type, "basic") == 0) {

    consumer = new Basic_Consumer(reference_reader,
				  kernel_trace_pathname,
				  output_pathname);

  } else if (strcmp(merge_type, "per-task") == 0) {

    consumer = new Per_Task_Consumer(reference_reader,
				     kernel_trace_pathname,
				     output_pathname);

  } else {

    fprintf(stderr, "main(): Unknown merge type %s\n", merge_type);
    exit(1);

  }

  // Ensure that the consumer was allocated.
  if (consumer == 0) {

    fprintf(stderr, "main(): Consumer allocation failed\n");
    exit(1);

  }

  // Merge the traces.
  consumer->merge();

  // Eliminate the consumer object.
  delete consumer;
  consumer = 0;

  // Eliminate the reference reader object.
  delete reference_reader;
  reference_reader = 0;

  // Exit successfully.
  return 0;

} // main
// ===================================================================
