/**
 * \file types-and-constants.hh
 * \author Scott F. H. Kaplan <sfkaplan@cs.amherst.edu>
 * \date January 2003
 * \brief The simple types and constants used in many modules and
 *        classes.
 **/



// ===================================================================
// Avoid multiple inclusion.
#if !defined (_TYPES_AND_CONSTANTS_HH)
#define _TYPES_AND_CONSTANTS_HH
// ===================================================================



// ===================================================================
// INCLUDES
#include <hash_map.h>
#include <list.h>
// ===================================================================



// ===================================================================
// FORWARD DECLARATIONS

struct Task;
struct Process;
// ===================================================================



// ===================================================================
// TYPES

/** \brief The type for the tag that indicates the type of a
 *         reference. */
typedef char tag_t;

/** \brief The type for a time counter. */
typedef unsigned long long int timestamp_t;

/** \brief The type for a virtual address. */
typedef unsigned long int virtual_address_t;

/** \brief The type for a reference length (bytes accessed). */
typedef unsigned short int length_t;

/**
 * \brief The type for a context (virtual address space) ID.
 *
 * This value is taken as the pointer to the top of the page table
 * that was active at the time of a reference.
 **/
typedef unsigned long int context_ID_t;

/**
 * \brief The type for a virtual page number.
 * \see PAGE_SIZE_ORDER
 *
 * This value is taken as some number of upper bits from a virtual
 * address.
 **/
typedef unsigned long int virtual_page_ID_t;

/** \brief The type for a canonical page number. */
typedef unsigned long int canonical_page_ID_t;

/** \brief The type for a task (thread) ID. */
typedef unsigned short int task_ID_t;

/** \brief The type for a process ID. */
typedef unsigned int process_ID_t;

/** \brief The type for an inode number, and thus a unique ID for a
 *         file. */
typedef unsigned long int inode_ID_t;

/** \brief The type for a device ID, uniquely identifying a device
 *         that holds files. */
typedef unsigned short int device_ID_t;

/** \brief The type of a IPC shared memory segment. */
typedef unsigned long int shm_ID_t;

/** \brief The type for an offset into a space, usually a virtual
 *         space. */
typedef unsigned long int offset_t;

/** \brief The type for an offset into a file. */
typedef unsigned long long int file_offset_t;

/** \brief The type for a marker that indicates the type of a page
 *         (anonymous, file mmap'ed, etc.) */
typedef char page_type_t;
// ===================================================================



// ===================================================================
// CONSTANTS

/** \brief The size of buffer spaces allocated for temporary work. */
#define BUFFER_SIZE 256

/** \brief The number of offset bits in a virtual address.
 *
 * Given a page size n, the log_{2}(n), a.k.a. the number of lower
 * bits of an address to eliminate to create a page number.
 **/
#define PAGE_SIZE_ORDER 12;
 
// Valid reference trace tags.
#define TAG_USER_READ                'r'
#define TAG_KERNEL_READ              'R'
#define TAG_USER_WRITE               'w'
#define TAG_KERNEL_WRITE             'W'
#define TAG_USER_INSTRUCTION_FETCH   'i'
#define TAG_KERNEL_EVENT             'K'
#define TAG_KERNEL_INSTRUCTION_FETCH 'I'

// Valid kernel trace tags.
#define TAG_SCHEDULE                 'S'
#define TAG_FORK                     'F'
#define TAG_EXEC                     'X'
#define TAG_EXIT                     'T'
#define TAG_CONTEXT_ASSIGNMENT       'A'
#define TAG_DUPLICATE_RANGE          'D'
#define TAG_MMAP_FILE                'M'
#define TAG_MMAP_ANONYMOUS           'm'
#define TAG_MUNMAP                   'U'
#define TAG_COMPLETE_UNMAP           'u'
#define TAG_IPC_SHMAT                's'
#define TAG_IPC_SHMDT                'd'
#define TAG_COW_UNMAP                'C'
#define TAG_BUFFER_ALLOCATION        'B'
#define TAG_BUFFER_DEALLOCATION      'b'
#define TAG_FILE_CACHE_ALLOCATION    'P'
#define TAG_FILE_CACHE_DEALLOCATION  'p'
#define TAG_FILE_OPEN                '('
#define TAG_FILE_CLOSE               ')'
#define TAG_FILE_READ                '<'
#define TAG_FILE_WRITE               '>'
#define TAG_FILE_DELETE              '-'
#define TAG_ACCEPT                   'a'
#define TAG_END_OF_TRACE             'e'

// Page types, used to know what is underlying each canonical page.

/**
 * \brief Marks a canonical page as ``anonymous''.
 *
 * An <i>anonymous page</i> is allocated to a process but not backed
 * by a named file.  Heap space is typically anonymous.
 **/
#define TYPE_ANONYMOUS                   'A'

/** \brief Marks a canonical page as part of the VFS cache. */
#define TYPE_FILE_CACHE                  'F'

/** \brief Marks a canonical page as part of an IPC shared memory
 *         segment. */
#define TYPE_IPC_SHM_SEGMENT             'S'

/**
 * \brief Marks a canonical page as the ``zero page''.
 *
 * The <i>zero page</i> is a shared page maintained by the kernel
 * which it maps as copy-on-write for new allocations, thus ensuring
 * that new virtual pages are zeroed.  There should be only one of
 * these in the canonical address space.
 **/
#define TYPE_ZERO_PAGE                   'Z'

/** \brief Marks a canonical page as a buffer cache page. */
#define TYPE_BUFFER_CACHE                'B'

/**
 * \brief The kernel's context ID.
 *
 * This value is specific to Linux, which uses the address 0x101000 as
 * the top of the kernel's page table.  Used to identify references
 * made by kernel tasks.
 **/
#define KERNEL_CONTEXT 0x101

/** \brief The kernel's process ID. */
#define KERNEL_PROCESS_ID 0

/** \brief The first page number of the kernel's virtual space that is
 *         shared with all tasks. (Linux specific) */
#define KERNEL_VIRTUAL_PAGE_OFFSET 0xc0000

/**
 * \brief The first page number of the kernel's canonical space that is
 *        shared with all tasks.
 *
 *  The upper 1 GB of the canonical space is reserved for the shared
 *  kernel space, and this is the first page of that range, and thus
 *  the kernel_virtual_page_offset maps to this page in the
 *  virtual->canonical mapping.
 **/
#define KERNEL_CANONICAL_PAGE_OFFSET 0xfffc0000
// ===================================================================



// ===================================================================
// STRUCTURES AND MAP TYPES

/** \brief A structure for uniquely identifying a virtual address
 *         within its virtual address space. */
struct virtual_page_s {

  context_ID_t context_ID;
  virtual_page_ID_t page_ID;

};

/** \brief A structure for a canonical page, storing both its identity
 *         and its metadata. */
struct canonical_page_s {

  canonical_page_ID_t page; /**< The page # in canonical space. */
  unsigned int reference_count; /**< Virtual mappings to this page. */ 
  page_type_t type; /**< The type of this page (anonymous, etc.). */
  shm_ID_t shm_ID; /**< The IPC shm ID for this page (if it is of that
		      type). */

};

/** \brief A structure to hold a unique identifier for a file. */
struct file_ID_s {

   device_ID_t major; /**< The major device number. */
   device_ID_t minor; /**< The minor device number. */
   inode_ID_t inode_ID; /**< The inode number on the given device. */

};

/** \brief A structure for comparing virtual page numbers used by
 *         tables indexed by virtual page number. */
struct equal_virtual_page_ID_s {
  bool operator() (const virtual_page_ID_t x,
		   const virtual_page_ID_t y) const {
    return (x == y);
  }
};

/** \brief A table type used to map virtual page IDs to canonical
 *         pages. */
typedef hash_map<virtual_page_ID_t,
		 canonical_page_s*,
		 hash<virtual_page_ID_t>,
		 equal_virtual_page_ID_s> virtual_to_canonical_map_t;

/** \brief A structure for comparing task IDs used by tables indexed
 *         by task ID. */
struct equal_task_ID_s {
  bool operator() (const task_ID_t x, const task_ID_t y) {
    return (x == y);
  }
};

/** \brief A structure for comparing context IDs used by tables
 *         indexed by context. */
struct equal_context_ID_s {
  bool operator() (const context_ID_t x, const context_ID_t y) const {
    return (x == y);
  }
};

/** \brief A structure for comparing file IDs used by tables indexed
 *         by file ID. */
struct equal_file_ID_s {
  bool operator() (const file_ID_s& x, const file_ID_s& y) const {
    return ((x.major == y.major) &&
	    (x.minor == y.minor) &&
	    (x.inode_ID == y.inode_ID));
  }
};

/** \brief A structure for comparing IPC shared memory segment IDs
 *         used by tables indexed by shm IDs. */
struct equal_shm_ID_s {
  bool operator() (const shm_ID_t x, const shm_ID_t y) const {
    return (x == y);
  }
};

/**
 * \brief A structure that provides a hashing function for file IDs.
 *
 * This function is required because file IDs are not scalar values
 * for which STL provides a default hashing operation.
 **/
struct file_ID_hash_s {

  size_t operator() (const file_ID_s& id) const {

    // Treat the value as a sequence of bytes, forming a hash from a
    // calculation borrowed from the strings hashing function in the
    // STL code itself.
    unsigned long int h = 0;
    char* s = (char*)&id;
    for (unsigned int i = 0; i < sizeof(file_ID_s); i++) {
      h = (5 * h) + *s;
      s++;
    }

    return (size_t)h;
  }

};

/** \brief A table type used to track the liveness of tasks, indexed
 *         by task ID. */
typedef hash_map<task_ID_t,
		 bool,
		 hash<task_ID_t>,
		 equal_task_ID_s> live_map_t;

/** \brief A table type used to map file IDs to virtual->canonical
 *         mapping tables. */
typedef hash_map<file_ID_s,
		 virtual_to_canonical_map_t*,
		 file_ID_hash_s,
		 equal_file_ID_s> file_ID_to_V2C_map_t;

/** \brief A table type used to map IPC shared memory segment IDs to
 *         virtual->canonical mapping tables. */
typedef hash_map<shm_ID_t,
		 virtual_to_canonical_map_t*,
		 hash<shm_ID_t>,
		 equal_shm_ID_s> shm_ID_to_V2C_map_t;

/** \brief A table type used to track IPC shared memory segments that
 *         become defunct and must be deleted. */
typedef hash_map<shm_ID_t,
		 bool,
		 hash<shm_ID_t>,
		 equal_shm_ID_s> defunct_shm_ID_map_t;

/** \brief A table type that maps task IDs to task objects. */
typedef hash_map<task_ID_t,
		 Task*,
		 hash<task_ID_t>,
		 equal_task_ID_s> task_ID_to_Task_map_t;

/** \brief A table type that maps contexts to task objects. */
typedef hash_map<context_ID_t,
		 Process*,
		 hash<context_ID_t>,
		 equal_context_ID_s> context_ID_to_Process_map_t;

/** \brief A structure to hold the contents of a reference trace
 *         record. */
struct reference_record_s {

  tag_t tag; /**< The type of reference. */
  timestamp_t cycle_timestamp; /**< The CPU cycle time of the
				  reference. */
  timestamp_t instruction_timestamp; /**< The instruction count at the
					time of the reference. */
  timestamp_t reference_timestamp; /**< The reference count at the
				      time of the reference. */
  length_t length; /**< The number of bytes accessed. */
  virtual_address_t virtual_address; /**< The virtual address
					referenced. */
  virtual_page_s virtual_page; /**< The virtual page referenced. */

};

/**
 * \brief A structure to hold the contents of a kernel trace record.
 *
 * Depending on the type of event represented, some fields may not be
 * used or may have slightly different uses.
 **/
struct kernel_record_s {

  tag_t tag; /**< The type of event. */
  timestamp_t cycle_timestamp; /**< The CPU cycle time of the
				  event. */
  task_ID_t task_ID;  /**< The task that caused or was the primary
			 target of the event. */
  task_ID_t parent_task_ID; /**< The task that requested a fork(),
			       where the task_ID field is the newly
			       forked process. */
  context_ID_t context_ID; /**< The context associated with the
			      event. */
  context_ID_t duplicate_context_ID; /**< For a context duplication,
					the destination context, where
					the previous field is the
					source context. */
  virtual_page_ID_t starting_page; /**< The first (and sometimes only)
				      virtual page for the event. */
  virtual_page_ID_t ending_page; /**< The last virtual page for those
				    events that specify a virtual
				    range. */
  file_ID_s file_ID; /**< The unique file ID associated with the
			event. */
  virtual_page_ID_t file_starting_page; /**< The first (and sometimes
					   only) page within the
					   file's address space. */
  virtual_page_ID_t file_ending_page; /**< The last page within the
					 file for events that specify
					 a range. */
  shm_ID_t shm_ID; /**< The IPC shared memory segment ID associated
		      with the event. */
  char filename[BUFFER_SIZE]; /**< The name of the file for this
				 operation.  Note that this identifier
				 is not unique as the file ID is, but
				 indicates the name by which the
				 process accessed the file. */

};
// ===================================================================



// ===================================================================
#endif // _TYPES_AND_CONSTANTS_HH
// ===================================================================
